import numpy as np

import device_inductance

from pytest import approx

from . import typical_outputs  # Required fixture

__all__ = ["typical_outputs"]


def test_circuits(typical_outputs: device_inductance.TypicalOutputs):
    """Circuit mutual inductances and tables are derived from coils,
    which are tested for physical consistency, so these tests
    focus on the differences between the coils and circuits
    instead of revisiting all the coil tests"""

    device = typical_outputs.device
    coils = device.coils
    circuits = device.circuits

    # Make sure all the coils are in a circuit exactly once
    all_coils_in_circuits = sum(
        [[c[0] for c in circuit.coils] for circuit in circuits], start=[]
    )

    for i in range(len(coils)):
        assert (
            i in all_coils_in_circuits
        ), f"Coil {i}, {device.get_coil_names()[i]}, missing from circuits"

    assert (
        len(set(all_coils_in_circuits)) == device.n_coils
    ), "Extra coils or missing coils in circuits"

    # Make sure the total self-inductance of the circuit system is positive
    # Even if every circuit is wound in reverse, the total should still be positive,
    # and we'll never have a system that bootstraps its own current to infinity.
    # That said, not every row or column will have a positive sum, only the whole matrix.
    assert (
        np.sum(device.circuit_mutual_inductances) > 0.0
    ), "Nonphysical circuit mutual inductance matrix"

    assert (
        np.sum(device.circuit_structure_mutual_inductances) > 0.0
    ), "Nonphysical circuit-structure mutual inductance matrix"

    #    Circuit-mode mutual inductances are really part of a larger matrix
    #    and don't necessarily have a positive sum on their own
    circuit_mode_total_inductance = 0.0
    circuit_mode_total_inductance += 2.0 * np.sum(
        device.circuit_structure_mode_mutual_inductances
    )
    circuit_mode_total_inductance += np.sum(device.circuit_mutual_inductances)
    circuit_mode_total_inductance += np.sum(device.structure_mode_mutual_inductances)
    assert (
        circuit_mode_total_inductance > 0.0
    ), "Nonphysical circuit mutual inductance matrix"

    # All circuit self-inductances should be positive for the same reason
    assert np.all(
        np.diag(device.circuit_mutual_inductances) > 0.0
    ), "Bad circuit self-inductance"

    # Circuit system total inductance should be less than or equal to coils,
    # since reversing any coils in a circuit will reduce total inductance
    assert np.sum(device.circuit_mutual_inductances) <= np.sum(
        device.coil_mutual_inductances
    ), "Nonphysical circuit mutual inductance matrix"

    # Circuits' resistances should match the sum of coils, since it's all series segments
    assert np.sum(device.circuit_resistances) == approx(np.sum(
        device.coil_resistances
    ), rel=1e-4), "Circuit resistance does not match sum of coils"

    # Make sure the circuit tables sums match the coils with signs
    psi_circ = device.circuit_flux_tables
    br_circ, bz_circ = device.circuit_flux_density_tables
    psi_c = device.coil_flux_tables
    br_c, bz_c = device.coil_flux_density_tables
    for i, circuit in enumerate(device.circuits):
        circuit_flux_sum = np.sum(psi_circ[i])
        circuit_br_sum = np.sum(br_circ[i])
        circuit_bz_sum = np.sum(bz_circ[i])

        coil_flux_sum = sum([sign * np.sum(psi_c[j]) for j, sign in circuit.coils])
        coil_br_sum = sum([sign * np.sum(br_c[j]) for j, sign in circuit.coils])
        coil_bz_sum = sum([sign * np.sum(bz_c[j]) for j, sign in circuit.coils])

        assert circuit_flux_sum == approx(
            coil_flux_sum, rel=1e-6
        ), "Circuit table does not match coils"
        assert circuit_br_sum == approx(
            coil_br_sum, rel=1e-6
        ), "Circuit table does not match coils"
        assert circuit_bz_sum == approx(
            coil_bz_sum, rel=1e-6
        ), "Circuit table does not match coils"



"""
Method for discretizing cross-section outlines of conducting structure.
Mainly copied from Christoph Hasse's work on early device description adapters for MEQ.
"""

import itertools
from enum import Enum, auto

from shapely import Polygon

import gmsh
import numpy as np
from numpy.typing import NDArray


class MeshMode(Enum):
    """Meshing Mode"""

    Quad = auto()
    QuasiStructuredQuad = auto()
    Triangular = auto()


def _mesh_region(
    rz: NDArray,
    min_length: float = 0.1,
    max_length: float = 0.3,
    mesh_mode: MeshMode = MeshMode.QuasiStructuredQuad,
) -> list[Polygon]:
    """
    Generate a structured 2D mesh from an outline.

    Args:
        rz_in: [m] (Nx2) Region outline coords, interleaved
        min_length: [m] target minimum edge length in mesh. Defaults to 0.1.
        max_length: [m] target maximum edge length in mesh. Defaults to 0.3.
        mesh_mode: algorithm for meshing. Defaults to MeshMode.QuasiStructuredQuad.

    Raises:
        ValueError: If an unhandled MeshMode is provided

    Returns:
        Mesh element polygons, with number of vertices depending on method
    """
    # Notes from original implementation:
    # use_qsq enable the quasi structured quadrilaterl meshing
    # produces mesh that has elements more square like but seems to not
    # generate very reproducible meshes. so for testing it's better to disable
    # we need to close polygon in gmsh by reusing the same gmsh point so no need for the last closing point
    if np.all(np.isclose(rz[0], rz[-1])):
        rz = rz[:-1]

    gmsh.initialize()

    pts = [gmsh.model.geo.add_point(p[0], p[1], 0) for p in rz]

    lines = [gmsh.model.geo.add_line(s, e) for (s, e) in itertools.pairwise(pts)]
    # and the closing one
    lines.append(gmsh.model.geo.add_line(pts[-1], pts[0]))
    loop = gmsh.model.geo.addCurveLoop(lines)
    surf = gmsh.model.geo.addPlaneSurface([loop])

    gmsh.model.geo.synchronize()

    gmsh.option.setNumber("General.Verbosity", 0)
    gmsh.option.setNumber("Mesh.Smoothing", 100)
    gmsh.option.setNumber("Mesh.CharacteristicLengthMin", min_length)
    gmsh.option.setNumber("Mesh.CharacteristicLengthMax", max_length)

    # Set up for nearly-deterministic meshing
    gmsh.option.setNumber("Mesh.RandomSeed", 12394871234)
    #  Note this random factor is smaller than recommended, but
    #  larger values result in behavior so non-repeatable that it causes
    #  frequent unit test failures
    gmsh.option.setNumber("Mesh.RandomFactor", 1e-60)
    gmsh.option.setNumber("Mesh.RandomFactor3D", 1e-60)
    gmsh.option.setNumber('General.NumThreads', 1)

    if mesh_mode == MeshMode.Triangular:
        gmsh.option.setNumber("Mesh.Algorithm", 8)
    elif mesh_mode == MeshMode.QuasiStructuredQuad:
        # we don't want to be left with any triangles! (see gmsh tutorial 11)
        gmsh.option.setNumber("Mesh.RecombinationAlgorithm", 0)
        gmsh.option.setNumber("Mesh.RecombineAll", 1)
        gmsh.option.setNumber("Mesh.Algorithm", 11)
        gmsh.option.setNumber("Mesh.QuadqsSizemapMethod", 0)
    elif mesh_mode == MeshMode.Quad:
        gmsh.option.setNumber("Mesh.RecombineAll", 1)
        gmsh.option.setNumber("Mesh.Algorithm", 8)
        gmsh.option.setNumber("Mesh.RecombinationAlgorithm", 2)
    else:
        raise ValueError(f"Unknown MeshMode {mesh_mode}")

    # We can then generate a 2D mesh...
    gmsh.model.mesh.generate(2)

    n_nodes = 3 if mesh_mode == MeshMode.Triangular else 4

    # elem_node_tags, holds n_nodes (4 if quada or 3 if triangle) unique numerical identifiers
    # that identify the nodes that make up a quadrilateral/triangular mesh element
    # the vector is originally n_nodes * N long
    _, _, elem_node_tags = gmsh.model.mesh.get_elements(2, surf)
    # reshape into N x n_nodes, and substract 1 from the node identifier (see below)
    elem_node_tags = np.asarray(elem_node_tags[0]).reshape((-1, n_nodes)) - 1

    # To get the coordinates for an element we need to look up the coordinates
    # for each node from the nodes storage.

    # node_tags are numerical unique identifiers from 1...N
    # node_coords is a vector of 3N length holding x,y,z for each node
    node_tags, node_coords, _ = gmsh.model.mesh.get_nodes(2, surf, includeBoundary=True)
    node_tags = np.asarray(node_tags)
    # turn it into N x 2 vector
    node_coords = np.asarray(node_coords).reshape(-1, 3)[:, :2]  # only need X,Y (R,Z)
    # we substract 1 because then we can use 0... N-1 as index into the sorted array
    node_tags -= 1  # will be an idx
    node_sorting = np.argsort(node_tags)

    # ensure that we really have indices from 0...N-1
    assert np.all(node_tags[node_sorting] == np.arange(len(node_tags)))
    # now our coordinates are sorted according to that array.
    # So looking up the coordinates of node  id=5 is node_coords[5]
    node_coords = node_coords[node_sorting]

    # this is now an array of shape (N, n_nodes, 2)
    # N elements, defined by n_nodes sets of (R,Z) coordinate pairs
    mesh_element_coords = node_coords[elem_node_tags]

    gmsh.finalize()

    mesh_polygons = [Polygon(x) for x in mesh_element_coords]

    return mesh_polygons



import numpy as np
from numpy.typing import NDArray
from pytest import approx
from device_inductance.contour import trace_contour


def test_contour():
    """Make sure the contour can trace a circle"""

    xgrid = np.linspace(1.0, 5.0, 100)
    ygrid = np.linspace(-7.0, 7.0, 100)
    xmesh, ymesh = np.meshgrid(xgrid, ygrid, indexing="ij")
    x0 = 3.0
    z = guess_psi(xmesh, ymesh, x0, z0=0.0)

    mask = np.ones_like(z)

    # Draw a contour that should be a circle around the axis
    radius = 0.5
    start = (x0 + radius, 0.0)
    maxis = (x0, 0.0)
    limiter_circumference = 7.0 + 7.0 + 4.0 + 4.0
    contour = trace_contour(
        (xgrid, ygrid), z, start, maxis, limiter_circumference, mask
    )

    # Make sure the resulting contour is centered around the known center
    assert contour is not None
    assert np.mean(contour[0]) == approx(x0, abs=1e-2)
    assert np.mean(contour[1]) == approx(0.0, abs=1e-2)

    # Make sure the resulting contour is within tolerance of the true circular contour
    dx = contour[0] - x0
    dy = contour[1]
    dist = (dx**2 + dy**2) ** 0.5
    expected_dist = radius * np.ones_like(dist)
    assert np.allclose(dist, expected_dist, atol=2e-3)


def guess_psi(
    rmesh: NDArray, zmesh: NDArray, r0: float, z0: float, elongation: float = 1.0
) -> NDArray:
    """
    Qualitative initial guess for a flux distribution, centered at R0.
    Does not produce a specific minor radius or any other particulars, just a smooth
    falloff from a centroid at (r0, z0). Does not represent any attempt at
    a self-consistent or physically-correct solution.
    """
    rmin = rmesh[0][0]
    rmax = rmesh[-1][0]
    rnorm = (rmesh - r0) / (rmax - rmin)  # not in [0, 1]
    znorm = (zmesh - z0) / (rmax - rmin)  # not in [0, 1], normalized to R span
    psi = np.exp(-((rnorm**2 + (znorm / elongation) ** 2) ** 0.5))
    return psi  # [Wb] qualitative smooth distribution



"""End-to-end workflow generating all outputs and some exploratory plots"""

from itertools import chain, cycle

import numpy as np
import matplotlib.pyplot as plt

import device_inductance

# Don't spam the terminal if this is running on a build server
show_prog = plt.get_backend().lower() != "agg"

# Set up a regular computational grid
# The extent defined here will be adjusted during init to respect the required resolution
dxgrid = (0.05, 0.05)  # [m] grid resolution
dr, dz = dxgrid
extent = (dr * 2.0, 4.5, -3.0, 3.0)

# Load the default device
ods = device_inductance.load_default_ods()

# Pre-compute the usual set of matrices and tables
print("Generating typical outputs\n")
typical_outputs = device_inductance.typical(ods, extent, dxgrid, show_prog=show_prog)

# Pull out the full-featured device object
# which can be used to generate anything not in the typical workflow outputs
device = typical_outputs.device

# Populate some tables which are not part of the typical outputs
# and will be calculated on first access
print("\nGenerating optional outputs\n")
psi_mesh_plasma = device.plasma_flux_tables
br_coils, bz_coils = device.coil_flux_density_tables
br_structures, bz_structures = device.structure_flux_density_tables
br_modes, bz_modes = device.structure_mode_flux_density_tables
m_circuit_circuit = device.circuit_mutual_inductances
m_circuit_structure = device.circuit_structure_mutual_inductances
m_circuit_mode = device.circuit_structure_mode_mutual_inductances
psi_circuits = device.circuit_flux_tables
br_circuits, bz_circuits = device.circuit_flux_density_tables

# Get the extent and mesh to use for plots
extent = typical_outputs.extent_for_plotting
ar = (extent[3] - extent[2]) / (extent[1] - extent[0])  # aspect ratio of mesh

# Make some validation plots
ncoil = device.n_coils
ncirc = device.n_circuits
nmodes = typical_outputs.nmodes


def table_imshow(arr, contours=True):
    """Helper function of plotting flux and B-field tables"""
    cmap = "seismic"
    vmax = np.max(np.abs(arr))
    vmin = -vmax
    plt.imshow(
        arr,
        extent=extent,
        origin="lower",
        interpolation="bicubic",
        cmap=cmap,
        vmin=vmin,
        vmax=vmax,
    )

    if contours:
        _cs = plt.contour(
            arr,
            extent=extent,
            levels=4,
            colors="k",
            linewidths=1,
        )


plt.figure(figsize=(5, 6.5))
structure_rs = list(chain(*[x.rs for x in device.structures]))
structure_zs = list(chain(*[x.zs for x in device.structures]))
mask_for_plot = 1.0 - device.limiter_mask.copy()
plt.imshow(
    device.limiter_mask.T,
    origin="lower",
    extent=extent,
    interpolation="nearest",
    cmap="BuGn",
)
plt.scatter(
    structure_rs,
    structure_zs,
    s=5,
    marker=".",
    color="k",
    alpha=1,
)
color_cycle = cycle(["#d10606", "#06d12b", "#06c7d1"])
for s in device.structures:
    plt.plot(*s.polygon.boundary.xy, linewidth=3, color=next(color_cycle))
    for f in s.filaments:
        plt.plot(*f.polygon.boundary.xy, linewidth=1, color='k', alpha=0.7)
    

for c in device.coils:
    coil_rs = [f.r for f in c.filaments]
    coil_zs = [f.z for f in c.filaments]
    plt.scatter(coil_rs, coil_zs, color="k")
plt.plot(*device.limiter.boundary.xy, color="k")
plt.gca().axis("equal")
plt.title("Filaments and Structure Mesh")
plt.xlabel("R [m]")
plt.ylabel("Z [m]")
plt.tight_layout()

# plt.figure()  # matshow makes its own figure
plt.matshow(typical_outputs.mcc)
plt.colorbar()
plt.title("Coil-Coil Mutual Inductances [H]")

plt.matshow(m_circuit_circuit)
plt.colorbar()
plt.title("Circuit-Circuit Mutual Inductances [H]")

plt.figure()
plt.imshow(
    m_circuit_structure,
    extent=[0, 1, 0, 1],
    origin="lower",
    interpolation="nearest",
)
plt.colorbar()
plt.title("Circuit-Structure Mutual Inductances [H]")

plt.matshow(m_circuit_mode)
plt.colorbar()
plt.title("Circuit-Mode Mutual Inductances [H]")

plt.matshow(typical_outputs.mss)
plt.colorbar()
plt.title("Passive Structure Mutual Inductances [H]")

plt.figure()
plt.imshow(
    typical_outputs.mcs,
    extent=[0, 1, 0, 1],
    origin="lower",
    interpolation="nearest",
)
plt.colorbar()
plt.title("Coil-Structure Mutual Inductances [H]")

plt.figure()
plt.imshow(
    typical_outputs.tuv, extent=[0, 1, 0, 1], origin="lower", interpolation="nearest"
)
d = device.structure_mode_eigenvalues
plt.plot(
    np.linspace(0.0, 1.0, nmodes, endpoint=True),
    np.abs(d / d[0]),
    color="k",
    label="Normalized Eigenvalues",
)  # eigenvalues
plt.legend()
plt.colorbar()
plt.title(f"Tuv, Structure Model Reduction Transform\n(neig = {nmodes})")

n = int(nmodes**0.5)
m = nmodes // n
n = n if n * m >= nmodes else n + 1
fig, axes = plt.subplots(n, m, figsize=(min(n, 8), min(m, 8)))
axes = sum([[y for y in x] for x in axes], start=[])
for i in range(nmodes):
    plt.sca(axes[i])
    table_imshow(
        typical_outputs.psi_modes[i].T,
    )
    plt.scatter(structure_rs, structure_zs, s=5, marker=".", color="k", alpha=1)
    plt.title(f"Mode {i}")
for ax in axes:
    plt.sca(ax)
    plt.axis("off")
plt.suptitle("Eigenmode Flux Maps")

n = int(nmodes**0.5)
m = nmodes // n
n = n if n * m >= nmodes else n + 1
fig, axes = plt.subplots(n, m, figsize=(min(n, 8), min(m, 8)))
axes = sum([[y for y in x] for x in axes], start=[])
for i in range(nmodes):
    plt.sca(axes[i])
    table_imshow(br_modes[i].T)
    plt.scatter(structure_rs, structure_zs, s=5, marker=".", color="k", alpha=1)
    plt.title(f"Mode {i}")
for ax in axes:
    plt.sca(ax)
    plt.axis("off")
plt.suptitle("Eigenmode Br Maps")

n = int(nmodes**0.5)
m = nmodes // n
n = n if n * m >= nmodes else n + 1
fig, axes = plt.subplots(n, m, figsize=(min(n, 8), min(m, 8)))
axes = sum([[y for y in x] for x in axes], start=[])
for i in range(nmodes):
    plt.sca(axes[i])
    table_imshow(bz_modes[i].T)
    plt.scatter(structure_rs, structure_zs, s=5, marker=".", color="k", alpha=1)
    plt.title(f"Mode {i}")
for ax in axes:
    plt.sca(ax)
    plt.axis("off")
plt.suptitle("Eigenmode Bz Maps")

plt.figure()
plt.title("Sum of Passive Filament Flux Maps\nNormalized By Area")
structure_areas = np.array([s.polygon.area for s in typical_outputs.device.structures])  # [m^2]
structure_areas = structure_areas.reshape((len(structure_areas), 1, 1))
structure_area_fracs = structure_areas / np.sum(structure_areas)
table_imshow(np.sum(typical_outputs.psi_s * structure_area_fracs, axis=0).T)
plt.scatter(
    structure_rs,
    structure_zs,
    s=5,
    marker=".",
    color="k",
    alpha=1,
)


fig, axes = plt.subplots(1, 2, sharex=True, sharey=True)
plt.suptitle("Sum of Passive Structure Filament B-field Maps\nNormalized By Area")
plt.sca(axes[0])
table_imshow(np.sum(br_structures * structure_area_fracs, axis=0).T)
plt.scatter(
    structure_rs,
    structure_zs,
    s=5,
    marker=".",
    color="k",
    alpha=1,
)
plt.title("Br")
plt.sca(axes[1])
table_imshow(np.sum(bz_structures * structure_area_fracs, axis=0).T)
plt.scatter(
    structure_rs,
    structure_zs,
    s=5,
    marker=".",
    color="k",
    alpha=1,
)
plt.title("Bz")


n = int(ncoil**0.5)
m = ncoil // n
n = n if n * m >= ncoil else n + 1
fig, axes = plt.subplots(n, m, figsize=(6 * n / m / ar, ar * 8 * m / n))
plt.suptitle("Coil Flux Maps")
axes = sum([[y for y in x] for x in axes], start=[])
for i in range(ncoil):
    coil = device.coils[i]
    plt.sca(axes[i])
    table_imshow(typical_outputs.psi_c[i].T)
    plt.scatter(
        [x.r for x in coil.filaments],
        [x.z for x in coil.filaments],
        marker=".",
        color="k",
        s=5,
    )
    plt.title(coil.name)
for ax in axes:
    plt.sca(ax)
    plt.axis("off")

n = int(ncoil**0.5)
m = ncoil // n
n = n if n * m >= ncoil else n + 1
fig, axes = plt.subplots(n, m, figsize=(6 * n / m / ar, ar * 8 * m / n))
plt.suptitle("Coil Br Maps")
axes = sum([[y for y in x] for x in axes], start=[])
for i in range(ncoil):
    coil = device.coils[i]
    plt.sca(axes[i])
    table_imshow(br_coils[i].T)
    plt.scatter(
        [x.r for x in coil.filaments],
        [x.z for x in coil.filaments],
        marker=".",
        color="k",
        s=5,
    )
    plt.title(coil.name)
for ax in axes:
    plt.sca(ax)
    plt.axis("off")

n = int(ncoil**0.5)
m = ncoil // n
n = n if n * m >= ncoil else n + 1
fig, axes = plt.subplots(n, m, figsize=(6 * n / m / ar, ar * 8 * m / n))
plt.suptitle("Coil Bz Maps")
axes = sum([[y for y in x] for x in axes], start=[])
for i in range(ncoil):
    coil = device.coils[i]
    plt.sca(axes[i])
    table_imshow(bz_coils[i].T)
    plt.scatter(
        [x.r for x in coil.filaments],
        [x.z for x in coil.filaments],
        marker=".",
        color="k",
        s=5,
    )
    plt.title(coil.name)
for ax in axes:
    plt.sca(ax)
    plt.axis("off")


n = int(ncirc**0.5)
m = ncirc // n
n = n if n * m >= ncirc else n + 1
fig, axes = plt.subplots(n, m, figsize=(6 * n / m / ar, ar * 8 * m / n))
plt.suptitle("Circuit Flux Maps")
axes = sum([[y for y in x] for x in axes], start=[])
for j in range(ncirc):
    plt.sca(axes[j])
    circuit = device.circuits[j]
    table_imshow(psi_circuits[j].T)
    for i in [c[0] for c in circuit.coils]:
        coil = device.coils[i]
        plt.scatter(
            [x.r for x in coil.filaments],
            [x.z for x in coil.filaments],
            marker=".",
            color="k",
            s=5,
        )
        plt.title(circuit.name)
for ax in axes:
    plt.sca(ax)
    plt.axis("off")


n = int(ncirc**0.5)
m = ncirc // n
n = n if n * m >= ncirc else n + 1
fig, axes = plt.subplots(n, m, figsize=(6 * n / m / ar, ar * 8 * m / n))
plt.suptitle("Circuit Br Maps")
axes = sum([[y for y in x] for x in axes], start=[])
for j in range(ncirc):
    plt.sca(axes[j])
    circuit = device.circuits[j]
    table_imshow(br_circuits[j].T)
    for i in [c[0] for c in circuit.coils]:
        coil = device.coils[i]
        plt.scatter(
            [x.r for x in coil.filaments],
            [x.z for x in coil.filaments],
            marker=".",
            color="k",
            s=5,
        )
        plt.title(circuit.name)
for ax in axes:
    plt.sca(ax)
    plt.axis("off")

n = int(ncirc**0.5)
m = ncirc // n
n = n if n * m >= ncirc else n + 1
fig, axes = plt.subplots(n, m, figsize=(6 * n / m / ar, ar * 8 * m / n))
plt.suptitle("Circuit Bz Maps")
axes = sum([[y for y in x] for x in axes], start=[])
for j in range(ncirc):
    plt.sca(axes[j])
    circuit = device.circuits[j]
    table_imshow(bz_circuits[j].T)
    for i in [c[0] for c in circuit.coils]:
        coil = device.coils[i]
        plt.scatter(
            [x.r for x in coil.filaments],
            [x.z for x in coil.filaments],
            marker=".",
            color="k",
            s=5,
        )
        plt.title(circuit.name)
for ax in axes:
    plt.sca(ax)
    plt.axis("off")

plt.show()



"""
Eigenmode-based dimensionality reduction for the part of the inductor
system that is a coupled L-R system - that is, conducting structure and
and coils that are shorted.
Shorted coils are taken as a kind of conducting structure here.

Suppose we have a symmetric NxN system mutual inductance matrix like this:
```
    Mutual Inductance
    Coils      Structure
 __________ _______________
|          |               |
|     0    |       2       | P
|          |               |
|__________|_______________|
|          |               |
|          |               |
|     3    |       1       | Q
|          |               |
|          |               |
|__________|_______________|
      P            Q
```

The regions are

* 0:   Coil-coil interaction, PxP
* 1:   Structure-Structure interaction QxQ
* 2,3: Coil-structure interaction, PxQ, QxP, with M_2.t() == M_3

...where P + Q = N.

We need to shrink regions 1,2,3 because while there are typically only
around 10 coils in region 0, there may be hundreds or thousands of
conducting structure elements contributing to the dimension of 1,2,3,
resulting in O(1e4)-O(1e6) nonzeros.

We'd rather have around O(1e2)-O(1e3) nonzeros in the entire
mutual inductance matrix, and we can sacrifice some modeling resolution
w.r.t. the conducting structure in order to achieve that dimensionality
reduction while preserving the detail of the active coils in region 0.

In order to do this, we need a non-square system transformation matrix
of shape KxQ that can reduce the QxQ region 1 to some smaller size QxK,
and can also reduce the PxQ,QxP regions 2,3 to PxK,KxP while preserving their
effect the overall system behavior w.r.t. the coil region 0 as well
as possible.

Eigenvector decomposition is a good candidate for this - the matrix
of eigenvectors preserves the behavior of the system. It also gives
a convenient ordering of the importance of each row and column
(the eigenvalues) in order to allow us to truncate the matrix
of eigenvectors and arrive at a truncated KxQ shape.

Now, it's not specifically the mutual inductance matrix itself
that we need to shrink, but the whole corresponding state-space
system. In this case, regions 1,2,3 (passive structure) relate to a
coupled inductive-resistive system, with self-interaction time scales
```
    A = -inv(M_1) @ R,                 # units of (1/s)
```
where
M_1 is the portion of M in region 1, and
R is the diagonal matrix of loop resistances associated with each
piece of conducting structure or shorted coil:

```
Structure
resistances
 _______________
| .             |
|    .          |
|       R       |  Q
|          .    |
|             . |
|_______________|
        Q
```

So, we can get the transformation matrix we need by taking the
top few eigenvectors of A:
```
    Tuv = V_sorted[:, :K]         # dimensionless, sampling K columns
```
where K is an empirically-tuned integer value with, hopefully, K << Q.

This gives a rectangular transformation matrix that fits our need:
```
 __________
|          |
|          |
|    Tuv   | Q
|          |
|          |
|__________|
      K
```
and can be used as an adapter between
the reduced-order model of the structure self-interaction represented by
```
    M1*  = Tuv' @ (M1 @ Tuv)       # Units of H, reduced-order structure-structure model
```
and the coupling regions 2,3
```
    M_2* = M_2 @ Tuv     # Units of H
    M_3* = M_3 @ Tuv
```
as well as tensors tabulating linear field contributions from
each structural element (flux, B-field, mutual inductance, etc):
```
    psi* = psi @ Tuv     # With psi tables shaped like (W, Q) for W grid points
    br*  = br  @ Tuv
    ...and so on.
```
Notably, the transformation of the resistances needs different treatment
because it's really a column vector that's been stretched into a diagonal matrix:
```
    R*   = Tuv' @ (R @ Tuv)    # Expands vector information in a similar way to outer(R, R)
```
As a final note, while the system for dI/dt uses the inverse of M, we really want the smallest
eigenvalues that would come out of that - the lowest frequency modes, which are most relevant
to controls. To avoid taking the (very ill-conditioned) inverse of M, we can instead solve the
system for `1/dI/dt = (-R^-1 @ M) @ v` and take the largest-magnitude eigenvalues with significantly
reduced numerical error.

This is not the _only_ reasonable choice for formulating the transformation matrix.
In particular, some other methods offer particular advantages in exchange for complexity:
    * Iterative SVD can be faster than proper eigenvalue decomposition
    * State-space balanced reduction gives a closer approximation of the important
      parts of system dynamics, especially phase response, but can suffer numerically with large systems
"""

from typing import Tuple

import numpy as np
from numpy.typing import NDArray


def eigenmode_reduction(
    m: NDArray, r: NDArray, max_neig: int | None
) -> Tuple[NDArray, NDArray, int]:
    """
    Do eigenmode decomposition of inductive-resistive system and truncate
    to top `max_neig` terms.

    See module-level docs for more detail about the model reduction approach.

    Args:
        m: [H]   QxQ symmetric mutual inductance matrix for conducting structure
        r: [ohm] QxQ diagonal resistance matrix for conducting structure
        max_neig: Optional maximum number of eigenvalues to keep. None -> keep all.

    Returns:
        (d, tuv, neig), Eigenvalues in [s], QxK transformation matrix with K <= Q, and number of terms
    """
    rvec = np.diag(r)
    r_inv = np.diag(1.0 / rvec)  # [ohm^-1]

    # Build system describing 1/d(current)/d(time) timescales
    # under zero applied voltage (balanced inductive-resistive decay)
    # with unit current.
    a = -r_inv @ m  # [s] Coupled inductive-resistive system timescales

    # Do eigendecomposition using method for asymmetric matrices.
    # Per tradition, we have eigenvalues `d` and eigenvectors `v`.
    d, v = np.linalg.eig(a)  # ([s], [dimensionless])

    # Sort eigenvalues by magnitude and keep the permutation,
    # then reorder to lead with largest terms
    inds = np.flip(np.argsort(d**2))  # Sort for largest magnitude eigenvalues first
    tuv = v[:, inds]
    d = d[inds]

    # Truncate to take just the highest-magnitude terms
    if max_neig is not None:
        tuv = v[:, :max_neig]
        d = d[:max_neig]

    # The actual number of terms retained, which may be less
    # than the maximum allowable if the size of the supplied
    # system was smaller than max_neig.
    neig = len(d)

    return d, tuv, neig


def stabilized_eigenmode_reduction(
    m: NDArray, r: NDArray, max_neig: int | None
) -> tuple[NDArray, NDArray, int]:
    """
    This method can be more computationally stable than the direct method,
    and may give better results for slightly-asymmetric systems that are
    intended to be fully physically symmetric.

    This is essentially direct (eigendecomposition) PCA.

    See module-level docs for more detail about the model reduction approach.

    Args:
        m: [H]   QxQ symmetric mutual inductance matrix for conducting structure
        r: [ohm] QxQ diagonal resistance matrix for conducting structure
        max_neig: Optional maximum number of eigenvalues to keep. None -> keep all.

    Returns:
        (d, tuv, neig), Eigenvalues in [s], QxK transformation matrix with K <= Q, and number of terms
    """
    # Unpack
    rvec = np.diag(r)  # [ohm] Extract diagonal
    r_inv = np.diag(1.0 / rvec)  # [ohm^-1]

    # Build system describing 1/(d(current)/d(time)) timescales
    # Then, take the covariance matrix to set up for direct PCA.
    #
    # To get a result with the correct units here, we have the option
    # to either take the elementwise square root now, or take the elementwise
    # square root of the solved eigenvalues. The correct choice is to take
    # the square root of the solved eigenvalues, because applying a nonlinear
    # function to the covariance matrix gives unfortunate effects including
    # nonphysical negative eigenvalues, and does not agree with the linear
    # algebra logic of PCA.
    a = -r_inv @ m  # [s] same system as the other method
    cov = a.T @ a  # [s^2] covariance matrix, symmetric positive semidefinite

    # Do eigendecomposition using method for Hermitian matrices
    # because this one is conveniently real and symmetric,
    # and this improves speed and numerical error.
    # Per tradition, we have eigenvalues `d` and eigenvectors `v`.
    d, v = np.linalg.eigh(cov)  # ([s^2], [dimensionless])
    #    Get the actual timescales instead of the square.
    #    These corrected eigenvalues are the "loadings" in PCA jargon.
    #    Some eigenvalues may come out slightly negative around 1e-20, but only due to numerical noise.
    assert np.all(d.round(decimals=16) >= 0.0), "Eigenvalues of a covariance matrix should be positive"
    d = np.sqrt(d)  # [s]

    # Sort eigenvalues by magnitude and keep the permutation,
    # then reorder to lead with largest terms
    inds = np.flip(np.argsort(d))
    d = d[inds]
    tuv = v[:, inds]

    # Truncate to take just the highest-magnitude terms
    if max_neig is not None:
        tuv = tuv[:, :max_neig]
        d = d[:max_neig]

    # The actual number of terms retained, which may be less
    # than the maximum allowable if the size of the supplied
    # system was smaller than max_neig.
    neig = len(d)

    # Make sure outputs are contiguous in memory to speed up access later
    d = np.ascontiguousarray(d)
    tuv = np.ascontiguousarray(tuv)

    return d, tuv, neig



from importlib.metadata import metadata
from typing import Literal
from pathlib import Path

__version__ = metadata(str(__package__))["Version"]
__author__ = metadata(str(__package__))["Author"]

from omas import ODS, load_omas_json

from device_inductance import model_reduction, contour, sensors, logging, mesh, structures
from device_inductance.device import DeviceInductance, TypicalOutputs
from device_inductance.coils import Coil, CoilFilament
from device_inductance.structures import PassiveStructureLoop
from device_inductance.logging import log, logger_is_set_up, logger_setup_default
from device_inductance.utils import calc_flux_density_from_flux, gradient_order4, flux_solver, solve_flux_axisymmetric


def load_default_ods() -> ODS:
    """
    Load an example ODS file in the format required by device_inductance.

    The example differs from real SPARC configurations in at least the following ways:
      * Coil number of turns and resistances are obfuscated
      * Actual magnetics sensors are replaced with mockup examples
      * The example description may be arbitrarily out-of-date, as it is not updated regularly
    """

    if not logger_is_set_up():
        logger_setup_default()

    log().warning("""
Loading example device description.
The example differs from real SPARC configurations in at least the following ways:
    * Coil number of turns and resistances are obfuscated
    * Actual magnetics sensors are replaced with mockup examples
    * The example description may be arbitrarily out-of-date, as it is not updated regularly
                  """)

    # NOTE: This should be rewritten to use importlib once omas supports loading raw text
    ods_filename = (
        Path(__file__).parent / "../examples/OS_SPARC_Device_Description.json"
    )
    with open(ods_filename) as f:
        ods = load_omas_json(f)

    # Add some dummy sensors to exercise the sensor functions
    #   Full flux loop
    ffloop = ods["magnetics.flux_loop.0"]
    ffloop["type.index"] = 1
    ffloop["name"] = "dummy_full_flux_loop"
    ffloop["position.0.r"] = 3.0
    ffloop["position.0.z"] = 0.0
    #   Partial flux loop
    pfloop = ods["magnetics.flux_loop.1"]
    pfloop["type.index"] = 2
    pfloop["name"] = "dummy_partial_flux_loop"
    pfloop["position.0.r"] = 3.0
    pfloop["position.0.phi"] = 0.0
    pfloop["position.0.z"] = 0.0
    pfloop["position.1.r"] = 3.3
    pfloop["position.1.phi"] = 0.25
    pfloop["position.1.z"] = 1.0
    #   Bpol probe
    bpcoil = ods["magnetics.b_field_pol_probe.0"]
    bpcoil["type.index"] = 2
    bpcoil["name"] = "dummy_poloidal_b_field_probe"
    bpcoil["position.r"] = 3.1
    bpcoil["position.phi"] = 0.0
    bpcoil["position.z"] = 0.5
    bpcoil["poloidal_angle"] = 0.7

    return ods


def typical(
    ods: ODS,
    extent: tuple[float, float, float, float] = (0.0, 0.0, 0.0, 0.0),
    dxgrid: tuple[float, float] = (0.0, 0.0),
    max_nmodes: int = 40,
    model_reduction_method: Literal["eigenmode", "stabilized eigenmode"] = "eigenmode",
    show_prog: bool = True,
    plasma_coil_force_method: Literal["tables", "mask"] = "mask",
    n_radial_slices: int = 30,
) -> TypicalOutputs:
    """
    Generate a typical set of outputs,
    notably excluding the plasma flux tables which usually require much more
    run time than the rest of the outputs combined.

    Note: during initialization, the extent of the computational grid may be
    adjusted to achieve the target spatial resolution. The adjusted extent
    will always bound the requested extent.

    Args:
        ods: An OMAS object in the format produced by device_description
        extent: [m] Extent of computational domain; adjusted during init
        dxgrid: [m] Spatial resolution of computational grid
        max_nmodes: Maximum number of structure modes to keep. Defaults to 40.
        show_prog: Whether to show terminal progress bars. Defaults to True.
        plasma_coil_force_method: Whether to interpolate B-field on the fully-realized mesh tables,
                                  or do direct filament calculations from points inside the limiter mask.
                                  Defaults to "mask", which is faster and uses less memory, but only includes
                                  nonzero entries inside the limiter, which requires a valid limiter geometry.
        n_radial_slices: Number of radial slices to use for chunking large structures. Each slice is centered
                            at the limiter centroid.

    Returns:
        A fully-computed set of matrices and tables covering the needs of a typical workflow
    """
    device = DeviceInductance(
        ods=ods,
        max_nmodes=max_nmodes,
        extent=extent,
        dxgrid=dxgrid,
        model_reduction_method=model_reduction_method,
        show_prog=show_prog,
        plasma_coil_force_method=plasma_coil_force_method,
        n_radial_slices=n_radial_slices,
    )

    out = TypicalOutputs(
        device=device,
        extent=device.extent,
        dxgrid=device.dxgrid,
        meshes=device.meshes,
        grids=device.grids,
        extent_for_plotting=device.extent_for_plotting,
        mcc=device.coil_mutual_inductances,
        mss=device.structure_mutual_inductances,
        mcs=device.coil_structure_mutual_inductances,
        r_s=device.structure_resistances,
        r_c=device.coil_resistances,
        r_modes=device.structure_mode_resistances,
        tuv=device.structure_model_reduction,
        nmodes=device.n_structure_modes,
        psi_c=device.coil_flux_tables,
        psi_s=device.structure_flux_tables,
        psi_modes=device.structure_mode_flux_tables,
    )
    return out


__all__ = [
    "DeviceInductance",
    "TypicalOutputs",
    "typical",
    "Coil",
    "CoilFilament",
    "PassiveStructureLoop",
    "load_default_ods",
    "calc_flux_density_from_flux",
    "gradient_order4",
    "flux_solver",
    "solve_flux_axisymmetric",

    "logging",
    "model_reduction",
    "contour",
    "sensors",
    "mesh",
    "structures",
]



import numpy as np
from pytest import approx

import device_inductance
from . import typical_outputs, typical_outputs_many_slices  # Required fixture

__all__ = ["typical_outputs", "typical_outputs_many_slices"]


def test_structure_invariants(
    typical_outputs: device_inductance.TypicalOutputs,
    typical_outputs_many_slices: device_inductance.TypicalOutputs,
):
    """
    Check that the total self-inductance, total resistance, and mode eigenvalues
    are not changing much with changing discretization coarseness.
    """
    d1 = typical_outputs.device
    d2 = typical_outputs_many_slices.device

    # Make sure they both end up with the same total section area
    area1 = np.sum([s.polygon.area for s in d1.structures])
    area2 = np.sum([s.polygon.area for s in d2.structures])

    assert area1 == approx(area2, rel=1e-3), "Section area changed with discretization"

    # Make sure the total stored energy for a given current density on the cross-section
    # is not changing significantly with changing discretization
    i1 = np.array([s.polygon.area for s in d1.structures])  # [A] Reference current for 1 A/m^2 current density
    i2 = np.array([s.polygon.area for s in d2.structures])
    m1 = d1.structure_mutual_inductances  # [H] mutual inductance matrix between structure loops
    m2 = d2.structure_mutual_inductances

    e1 = 0.5 * i1.T @ m1 @ i1  # [J] stored energy for 1A/m^2 on the section
    e2 = 0.5 * i2.T @ m2 @ i2

    assert e1 == approx(e2, rel=1e-3), "Stored energy changed with discretization"

    # Make sure the parallel resistance is not changing significantly with changing discretization
    r1 = 1.0 / np.sum(1.0 / np.diag(d1.structure_resistances))  # [ohm]
    r2 = 1.0 / np.sum(1.0 / np.diag(d2.structure_resistances))  # [ohm]
    assert r1 == approx(r2, rel=1e-3), "Section resistance changed with discretization"

    # Make sure the top few structure model reduction eigenvalues are not changing much...
    eigd1 = d1.structure_mode_eigenvalues[:40]
    eigd1 = np.abs(eigd1 / eigd1[0])  # Normalize
    eigd2 = d2.structure_mode_eigenvalues[:40]
    eigd2 = np.abs(eigd2 / eigd2[0])

    assert np.allclose(eigd1, eigd2, rtol=0.01, atol=0.02), "Eigenvalue profile changed with discretization"



"""Tabulation of flux & related values on a provided computational grid"""

from itertools import product

import numpy as np
from numpy.typing import NDArray

from device_inductance.coils import Coil
from device_inductance.circuits import CoilSeriesCircuit
from device_inductance.structures import PassiveStructureLoop
from device_inductance.utils import (
    _progressbar,
    calc_flux_density_from_flux,
    _rect_mask,
)

from cfsem import (
    flux_circular_filament,
    flux_density_circular_filament,
    self_inductance_lyle6,
)

_MIN_DIST = 0.15  # [m]
"""
Under this distance from a filament, use a finite-difference calc on the
flux field to get the B-field instead of using a filament calc for B, because
the flux calcs are numerically better-conditioned.
"""


def _table_mask(
    meshes: tuple[NDArray, NDArray], extent: tuple[float, float, float, float]
) -> tuple[NDArray, tuple[NDArray[np.intp], ...]]:
    """Get a mask with padding of max(2*dx, _MIN_DIST) on each axis"""
    rmesh, zmesh = meshes
    dr = rmesh[1, 0] - rmesh[0, 0]  # [m]
    dz = zmesh[0, 1] - zmesh[0, 0]  # [m]
    rdelta = max(2 * dr, _MIN_DIST)  # [m]
    zdelta = max(2 * dz, _MIN_DIST)  # [m]
    return _rect_mask(meshes, extent, pad=(rdelta, zdelta))


def _calc_coil_flux_tables(
    coils: list[Coil], meshes: tuple[NDArray, NDArray], show_prog: bool = True
) -> NDArray:
    # Unpack
    ncoil = len(coils)
    rmesh, zmesh = meshes  # [m]
    shape = meshes[0].shape
    nr, nz = meshes[0].shape

    # Calculate tables using filament calcs
    coil_table_shape = (ncoil, nr, nz)  # This ordering makes each table contiguous
    psi_mesh_coils = np.zeros(coil_table_shape)  # [Wb/A]
    items = [x for x in enumerate(coils)]
    if show_prog:
        items = _progressbar(items, "Coil flux tables")
    for i, c in items:
        # Add contribution from each coil to its place in the table
        ifil = np.array([e.n for e in c.filaments])  # Effective current is nturns * 1A
        rfil = np.array([e.r for e in c.filaments])  # [m]
        zfil = np.array([e.z for e in c.filaments])  # [m]
        psi_mesh_coils[i, :, :] = flux_circular_filament(
            ifil, rfil, zfil, rmesh.flatten(), zmesh.flatten()
        ).reshape(shape)  # [Wb/A]

        # For coils with their winding pack on a regular grid, patch over the field near the coil
        # with a local axisymmetric flux solve
        cgrids = c.grids
        cmeshes = c.meshes
        cinterp = c.local_field_interpolators
        if cgrids is not None and cmeshes is not None and cinterp is not None:
            # Figure out which points to replace with the local field
            cpsi, _, _ = cinterp
            extent = c.extent
            _, inds = _table_mask(meshes, extent)

            # Replace points with local field solve
            robs = rmesh[inds].flatten()
            zobs = zmesh[inds].flatten()
            psi_mesh_coils[i, :, :][inds] = cpsi.eval([robs, zobs])

    return np.ascontiguousarray(psi_mesh_coils)  # [Wb/A]


def _calc_coil_flux_density_tables(
    coils: list[Coil],
    meshes: tuple[NDArray, NDArray],
    coil_flux_tables: NDArray,
    show_prog: bool = True,
) -> tuple[NDArray, NDArray]:
    # Unpack
    ncoil = len(coils)
    rmesh, zmesh = meshes  # [m]
    shape = meshes[0].shape
    nr, nz = meshes[0].shape

    # Calculate
    coil_table_shape = (ncoil, nr, nz)  # This ordering makes each table contiguous
    br_mesh_coils = np.zeros(coil_table_shape)  # [T/A]
    bz_mesh_coils = np.zeros(coil_table_shape)  # [T/A]
    items = [x for x in enumerate(coils)]
    if show_prog:
        items = _progressbar(items, "Coil flux density (B-field) tables")
    for i, c in items:
        # Add contribution from each coil to its place in the table
        ifil = np.array([e.n for e in c.filaments])  # Effective current is nturns * 1A
        rfil = np.array([e.r for e in c.filaments])  # [m]
        zfil = np.array([e.z for e in c.filaments])  # [m]
        b = flux_density_circular_filament(
            ifil, rfil, zfil, rmesh.flatten(), zmesh.flatten()
        )  # [T/A]
        br_mesh_coils[i, :, :] = b[0].reshape(shape)  # [T/A]
        bz_mesh_coils[i, :, :] = b[1].reshape(shape)  # [T/A]

        # Because the flux density filament calc has an additional 1/r^2 factor compared to
        # the flux calc, it is not as well-conditioned numerically and will tend to give
        # worse results for observation points very close to the coil filaments.
        # In order to remedy this, we can patch over the region immediately near the coil
        # winding pack using a B-field calc that comes from the flux calc, giving better
        # floating-point error at the expense of some discretization error.
        #    Figure out what part we're replacing
        _, inds = _table_mask(meshes, c.extent)
        #    Do the replacement
        br_from_psi, bz_from_psi = calc_flux_density_from_flux(
            coil_flux_tables[i, :, :], rmesh, zmesh
        )  # [T/A]
        br_mesh_coils[i][inds] = br_from_psi[inds]  # [T/A]
        bz_mesh_coils[i][inds] = bz_from_psi[inds]  # [T/A]

    br_mesh_coils = np.ascontiguousarray(br_mesh_coils)  # [T/A]
    bz_mesh_coils = np.ascontiguousarray(bz_mesh_coils)  # [T/A]

    return (br_mesh_coils, bz_mesh_coils)  # [T/A]


def _calc_structure_flux_tables(
    structures: list[PassiveStructureLoop],
    meshes: tuple[NDArray, NDArray],
    show_prog: bool = True,
) -> NDArray:
    # Unpack
    npassive = len(structures)
    rmesh, zmesh = meshes  # [m]
    shape = meshes[0].shape
    nr, nz = meshes[0].shape

    # Calculate
    structure_table_shape = (npassive, nr, nz)
    psi_mesh_structures = np.zeros(structure_table_shape)  # [Wb/A]
    items = [x for x in enumerate(structures)]
    show_every = max(1, npassive // 100)  # Don't spam too much
    if show_prog:
        items = _progressbar(items, "Structure flux tables", show_every)
    for i, s in items:
        # Add contribution from each structure filament to its place in the table
        #   [A] unit reference current for normalization times number of turns
        ifil = s.ns
        rfil = s.rs  # [m]
        zfil = s.zs  # [m]
        psi_mesh_structures[i, :, :] = flux_circular_filament(
            ifil, rfil, zfil, rmesh.flatten(), zmesh.flatten()
        ).reshape(shape)  # [Wb/A]

    return np.ascontiguousarray(psi_mesh_structures)  # [Wb/A]


def _calc_structure_flux_density_tables(
    structures: list[PassiveStructureLoop],
    meshes: tuple[NDArray, NDArray],
    structure_flux_tables: NDArray,
    show_prog: bool = True,
) -> tuple[NDArray, NDArray]:
    # Unpack
    npassive = len(structures)
    rmesh, zmesh = meshes  # [m]
    shape = meshes[0].shape
    nr, nz = meshes[0].shape

    # Calculate
    structure_table_shape = (npassive, nr, nz)
    br_mesh_structures = np.zeros(structure_table_shape)  # [T/A]
    bz_mesh_structures = np.zeros(structure_table_shape)  # [T/A]
    items = [x for x in enumerate(structures)]
    show_every = max(1, npassive // 100)  # Don't spam too much
    if show_prog:
        items = _progressbar(
            items, "Structure flux density (B-field) tables", show_every
        )
    for i, s in items:
        # Add contribution from each structure filament to its place in the table
        ifil = (
            s.ns
        )  # [A] unit reference current for normalization times number of turns
        rfil = s.rs  # [m]
        zfil = s.zs  # [m]
        b = flux_density_circular_filament(
            ifil, rfil, zfil, rmesh.flatten(), zmesh.flatten()
        )  # [T/A]
        br_mesh_structures[i, :, :] = b[0].reshape(shape)  # [T/A]
        bz_mesh_structures[i, :, :] = b[1].reshape(shape)  # [T/A]

        # Similar to the coils, we'll get better results very close to the
        # filaments using a finite difference on the flux values.
        # Another option would be to make a loop of linear segments and do biot-savart

        # Replacement table for this structure loop
        br_from_psi, bz_from_psi = calc_flux_density_from_flux(
            structure_flux_tables[i, :, :], rmesh, zmesh
        )  # [T/A]

        # Figure out what part we're replacing relative to each filament
        for r, z in zip(rfil, zfil):
            dist = ((rmesh - r) ** 2 + (zmesh - z) ** 2) ** 0.5  # [m]
            inds = np.where(dist < _MIN_DIST)

            br_mesh_structures[i, :, :][inds] = br_from_psi[inds]  # [T/A]
            bz_mesh_structures[i, :, :][inds] = bz_from_psi[inds]  # [T/A]

    br_mesh_structures = np.ascontiguousarray(br_mesh_structures)  # [T/A]
    bz_mesh_structures = np.ascontiguousarray(bz_mesh_structures)  # [T/A]

    return br_mesh_structures, bz_mesh_structures  # [T/A]


def _calc_mesh_flux_tables(
    meshes: tuple[NDArray, NDArray],
    grids: tuple[NDArray, NDArray],
    show_prog: bool = True,
) -> NDArray:
    # Unpack
    rgrid, zgrid = grids  # [m]
    rmesh, zmesh = meshes
    shape = meshes[0].shape
    nr, nz = shape
    dr = rgrid[1] - rgrid[0]
    dz = zgrid[1] - zgrid[0]
    assert np.allclose(np.diff(rgrid), dr, atol=1e-6), (
        "Self-inductance calc requires uniform grid"
    )
    assert np.allclose(np.diff(zgrid), dz, atol=1e-6), (
        "Self-inductance calc requires uniform grid"
    )

    # Calculate
    mesh_table_shape = (nr * nz, nr, nz)
    psi_mesh_mesh = np.zeros(mesh_table_shape)  # [Wb/A]
    items = [x for x in enumerate(product(range(nr), range(nz)))]
    show_every = max(1, (nr * nz) // 100)  # Don't spam too much
    if show_prog:
        items = _progressbar(items, "Mesh flux tables", show_every)
    ifil = np.array([1.0])  # [A] unit reference current for normalization
    rfil = np.array([np.nan])
    zfil = np.array([np.nan])
    for i, (ir, iz) in items:
        # For each grid location, calculate its contribution to all grid locations
        # including its own, using the axisymmetric filament calc for the non-self terms
        # and falling back on the rectangular-section calc for the self-term to resolve
        # the singularity.
        ifil = np.array([1.0])
        rfil = rgrid[ir]
        zfil = zgrid[iz]
        psi_mesh_mesh[i, :, :] = flux_circular_filament(
            ifil, rfil, zfil, rmesh.flatten(), zmesh.flatten()
        ).reshape(shape)
        # Replace singular self-term with 6th order rectangular-section calc
        psi_mesh_mesh[i, ir, iz] = self_inductance_lyle6(
            float(rgrid[ir]), float(dr), float(dz), n=1.0
        )

    return np.ascontiguousarray(psi_mesh_mesh)  # [Wb/A]


def _calc_structure_mode_flux_tables(
    psi_mesh_structures: NDArray,
    tuv: NDArray,
    show_prog: bool = True,
) -> NDArray:
    # Unpack
    neig = tuv.shape[1]
    npassive = psi_mesh_structures.shape[0]
    shape = psi_mesh_structures[0, :, :].shape
    nr, nz = shape

    # Calculate
    #    Eigenmode maps are a linear combination of
    #    the flux maps of the structure filaments
    eig_table_shape = (neig, nr, nz)
    psi_mesh_eig = np.zeros(eig_table_shape)  # [Wb/A]
    items = [x for x in product(range(neig), range(npassive))]
    if show_prog:
        items = _progressbar(items, "Structure mode flux tables")
    for i, j in items:
        psi_mesh_eig[i, :, :] += tuv[j, i] * psi_mesh_structures[j]

    return np.ascontiguousarray(psi_mesh_eig)


def _calc_structure_mode_flux_density_tables(
    br_mesh_structures: NDArray,
    bz_mesh_structures: NDArray,
    tuv: NDArray,
    show_prog: bool = True,
) -> tuple[NDArray, NDArray]:
    # Unpack
    neig = tuv.shape[1]
    npassive = br_mesh_structures.shape[0]
    shape = br_mesh_structures[0, :, :].shape
    nr, nz = shape

    # Calculate
    #    Eigenmode maps are a linear combination of
    #    the maps of the structure filaments
    eig_table_shape = (neig, nr, nz)
    br_mesh_eig = np.zeros(eig_table_shape)  # [T/A]
    bz_mesh_eig = np.zeros(eig_table_shape)  # [T/A]
    items = [x for x in product(range(neig), range(npassive))]
    if show_prog:
        items = _progressbar(
            items,
            "Structure mode flux density (B-field) table contributions",
            show_every=50,
        )
    for i, j in items:
        br_mesh_eig[i, :, :] += tuv[j, i] * br_mesh_structures[j]
        bz_mesh_eig[i, :, :] += tuv[j, i] * bz_mesh_structures[j]

    br_mesh_eig = np.ascontiguousarray(br_mesh_eig)
    bz_mesh_eig = np.ascontiguousarray(bz_mesh_eig)

    return br_mesh_eig, bz_mesh_eig  # [T/A]


def _calc_circuit_flux_tables(
    circuits: list[CoilSeriesCircuit], coil_flux_tables: NDArray, show_prog: bool = True
) -> NDArray:
    # Unpack
    ncirc = len(circuits)
    _, nr, nz = coil_flux_tables.shape

    # Sum coil contributions to each circuit response,
    # applying the series or anti-series orientation of each coil
    psi_circ = np.zeros((ncirc, nr, nz))  # [Wb/A]
    items = [x for x in enumerate(circuits)]
    if show_prog:
        items = _progressbar(items, "Circuit flux tables")
    for i, circuit in items:
        for coil_index, sign in circuit.coils:
            psi_circ[i, :, :] += sign * coil_flux_tables[coil_index, :, :]

    return psi_circ  # [Wb/A]


def _calc_circuit_flux_density_tables(
    circuits: list[CoilSeriesCircuit],
    br_mesh_coils: NDArray,
    bz_mesh_coils: NDArray,
    show_prog: bool = True,
) -> tuple[NDArray, NDArray]:
    # Unpack
    ncirc = len(circuits)
    _, nr, nz = br_mesh_coils.shape

    # Sum coil contributions to each circuit response,
    # applying the series or anti-series orientation of each coil
    br_circ = np.zeros((ncirc, nr, nz))  # [T/A]
    bz_circ = np.zeros((ncirc, nr, nz))  # [T/A]
    items = [x for x in enumerate(circuits)]
    if show_prog:
        items = _progressbar(items, "Circuit flux density (B-field) tables")
    for i, circuit in items:
        for coil_index, sign in circuit.coils:
            br_circ[i, :, :] += sign * br_mesh_coils[coil_index, :, :]
            bz_circ[i, :, :] += sign * bz_mesh_coils[coil_index, :, :]

    return br_circ, bz_circ  # [T/A]



from __future__ import annotations

from typing import Literal, Callable
from dataclasses import dataclass
from functools import cached_property

import numpy as np
from numpy.typing import NDArray

from cfsem import (
    self_inductance_distributed_axisymmetric_conductor,
)

from omas import ODS

from shapely import Polygon, Point

from device_inductance.structures import PassiveStructureLoop, _extract_structures
from device_inductance.coils import Coil, _extract_coils
from device_inductance.circuits import CoilSeriesCircuit, _extract_circuits
from device_inductance.sensors import (
    PoloidalFieldProbe,
    FullFluxLoop,
    PartialFluxLoop,
    _extract_full_flux_loops,
    _extract_partial_flux_loops,
    _extract_poloidal_field_probes,
)
from device_inductance.mutuals import (
    _calc_coil_mutual_inductances,
    _calc_structure_mutual_inductances,
    _calc_coil_structure_mutual_inductances,
    _calc_circuit_mutual_inductances,
    _calc_circuit_structure_mutual_inductances,
)
from device_inductance.tables import (
    _calc_coil_flux_tables,
    _calc_coil_flux_density_tables,
    _calc_structure_flux_tables,
    _calc_structure_flux_density_tables,
    _calc_structure_mode_flux_tables,
    _calc_structure_mode_flux_density_tables,
    _calc_mesh_flux_tables,
    _calc_circuit_flux_tables,
    _calc_circuit_flux_density_tables,
)
from device_inductance.forces import (
    _calc_coil_coil_forces,
    _calc_circuit_coil_forces,
    _calc_structure_coil_forces,
    _calc_plasma_coil_forces,
)
from device_inductance.utils import (
    calc_flux_density_from_flux,
    flux_solver,
    solve_flux_axisymmetric,
    _join_extents,
    _pad_extent,
)
from device_inductance import model_reduction
from device_inductance.logging import log, logger_is_set_up, logger_setup_default

F64 = np.float64


class DeviceInductance:
    """
    Thin wrapper to provide methods and properties on a device description ODS.
    Extracts geometries and calculates inductance matrices as well as flux fields
    and B-fields on the generated regular grid.

    Note:
        The contents of an instance of this class must be treated as
        immutable, as it relies on caching results to eliminate repeated
        computations along different paths through the analysis. Do not
        modify the supplied ODS after initialization or modify any of the
        values returned from methods or properties of this class!
    """

    _ods: ODS
    """OMAS data object in the format produced by device_description"""
    _max_nmodes: int = 40
    """Maximum number of structure modes to retain"""
    _min_extent: tuple[float, float, float, float] = (0.0, 0.0, 0.0, 0.0)
    """
    [m] rmin, rmax, zmin, zmax extent of computational domain.
    This will be updated during mesh initialization, during which it
    may be adjusted to satisfy the required spatial resolution.
    """
    _dxgrid: tuple[float, float] = (0.05, 0.05)
    """[m] spatial resolution of computational grid"""
    _model_reduction_method: Literal["eigenmode", "stabilized eigenmode"] = "eigenmode"
    """Choice of method for truncating passive structure system modes"""
    _show_prog: bool = False
    """Whether to display terminal progress bars during expensive calculations"""
    _plasma_coil_force_method: Literal["tables", "mask"] = "mask"
    """Whether to interpolate B-field on the fully-realized mesh tables,
    or do direct filament calculations from points inside the limiter mask.
    Defaults to "mask", which is faster and uses less memory, but only includes
    nonzero entries inside the limiter, which requires a valid limiter geometry."""
    _n_radial_slices: int = 30
    """Number of radial slices to use for chunking large structures. Each slice is centered
    at the limiter centroid."""

    def __init__(
        self,
        ods: ODS,
        *,
        show_prog: bool = False,
        max_nmodes: int = 40,
        min_extent: tuple[float, float, float, float] = (0.0, 0.0, 0.0, 0.0),
        dxgrid: tuple[float, float] = (0.05, 0.05),
        model_reduction_method: Literal[
            "eigenmode", "stabilized eigenmode"
        ] = "eigenmode",
        plasma_coil_force_method: Literal["tables", "mask"] = "mask",
        n_radial_slices: int = 30,
        **kwargs,  # For backwards compatibility with `extent` kwarg only
    ):
        """
        Attach a new DeviceInductance instance to an existing ODS description.

        Args:
            ods: OMAS data object in the format produced by device_description
            show_prog: Whether to display terminal progress bars during expensive calculations
            max_nmodes: Maximum number of structure modes to retain
            min_extent: [m] rmin, rmax, zmin, zmax extent of computational domain.
                    This will be updated during mesh initialization, during which it
                    may be adjusted to satisfy the required spatial resolution.
            dxgrid: [m] spatial resolution of computational grid
            plasma_coil_force_method: Whether to interpolate B-field on the fully-realized mesh tables,
                                      or do direct filament calculations from points inside the limiter mask.
                                      Defaults to "mask", which is faster and uses less memory, but only includes
                                      nonzero entries inside the limiter, which requires a valid limiter geometry.
            n_radial_slices: Number of radial slices to use for chunking large structures. Each slice is centered
                             at the limiter centroid.
        """
        if not logger_is_set_up():
            logger_setup_default()

        if "extent" in kwargs.keys():
            # Backwards compatibility with `extent` kwarg name only
            min_extent = kwargs.pop("extent")

        if len(kwargs) != 0:
            log().warning(f"DeviceInductance init ignoring extra kwargs: {kwargs}")

        self._ods = ods
        self._max_nmodes = max_nmodes
        self._min_extent = min_extent
        self._dxgrid = dxgrid
        self._model_reduction_method = model_reduction_method
        self._show_prog = show_prog
        self._plasma_coil_force_method = plasma_coil_force_method
        self._n_radial_slices = n_radial_slices

        self.__post_init__()

    def __hash__(self) -> int:
        # Do not need to hash immutable inputs
        return hash(id(self))

    def __post_init__(self):
        # Set a sensible default that encompasses the coils with 0.1m pad if none was provided
        if self._min_extent == (0.0, 0.0, 0.0, 0.0):
            inf = float("inf")
            self._min_extent = (inf, -inf, inf, -inf)
            for c in self.coils:
                self._min_extent = _join_extents(self._min_extent, c.extent)
            self._min_extent = _pad_extent(self._min_extent, (0.1, 0.1))
            # Don't auto-place grid cells too close to R=0
            rmin, rmax, zmin, zmax = self._min_extent
            rmin = max(rmin, 0.1)
            self._min_extent = (rmin, rmax, zmin, zmax)

        # Set a sensible default grid resolution to avoid dividing by zero
        if self._dxgrid == (0.0, 0.0):
            self._dxgrid = (0.05, 0.05)

        # Immutable after init, except for new cache entries
        def setattr_err(*_, **__):
            raise NotImplementedError(
                "DeviceInductance attributes are not intended to be mutated"
            )

        self.__setattr__ = setattr_err

    @property
    def ods(self) -> ODS:
        """
        The ODS description object with the device description.
        This object must not be edited after the DeviceInductance object
        is initialized!
        """
        return self._ods

    @property
    def max_nmodes(self) -> int:
        """Maximum number of (eigen)modes to keep in the passive structure model reduction"""
        return self._max_nmodes

    @property
    def min_extent(self) -> tuple[float, float, float, float]:
        """[m] (rmin, rmax, zmin, zmax) minimum extent requested at init.
        The actual extent of the mesh bounds this minimum, while respecting the
        required grid resolution exactly.
        """
        return self._min_extent

    @property
    def dxgrid(self) -> tuple[float, float]:
        """[m] (dr, dz) grid cell spacing"""
        return self._dxgrid

    @property
    def model_reduction_method(self) -> Literal["eigenmode", "stabilized eigenmode"]:
        """Choice of passive structure state-space model dimensionality reduction method"""
        return self._model_reduction_method

    @property
    def show_prog(self) -> bool:
        """Whether terminal progressbar will be displayed during some calcs"""
        return self._show_prog

    @cached_property
    def coils(self) -> list[Coil]:
        """Info about coil name and filaments"""
        return _extract_coils(self.ods)

    @cached_property
    def structures(self) -> list[PassiveStructureLoop]:
        """Info about structure filaments"""
        return _extract_structures(
            self.ods,
            limiter=self.limiter,
            n_slices=self._n_radial_slices,
            extent=self.extent_for_plotting,
            show_prog=self.show_prog,
        )

    @cached_property
    def circuits(self) -> list[CoilSeriesCircuit]:
        """Info about coils wired in series"""
        return _extract_circuits(self.ods)

    @cached_property
    def limiter(self) -> Polygon:
        """[m] Plasma-facing first-wall contour defined as a closed path around the plasma"""
        # This one is pretty quick, doesn't need its own file
        limiter_path_r = self.ods["wall.description_2d.0.limiter.unit.0.outline.r"]
        limiter_path_z = self.ods["wall.description_2d.0.limiter.unit.0.outline.z"]
        return Polygon(zip(limiter_path_r, limiter_path_z))

    @cached_property
    def poloidal_field_probes(self) -> list[PoloidalFieldProbe]:
        """List of poloidal B-field probe sensors"""
        return _extract_poloidal_field_probes(self.ods)

    @cached_property
    def full_flux_loops(self) -> list[FullFluxLoop]:
        """List of full flux loop sensors"""
        return _extract_full_flux_loops(self.ods)

    @cached_property
    def partial_flux_loops(self) -> list[PartialFluxLoop]:
        """List of partial flux loop sensors"""
        return _extract_partial_flux_loops(self.ods)

    @cached_property
    def limiter_mask(self) -> NDArray[F64]:
        """Mask with shape (nr, nz) indicating which parts of the mesh are inside the limiter"""
        limiter = self.limiter
        rmesh, zmesh = self.meshes
        nr, nz = rmesh.shape
        mask = np.ones_like(rmesh)
        # This is slow, but shapely only accepts one
        # point at a time
        for i in range(nr):
            for j in range(nz):
                r = rmesh[i, j]
                z = zmesh[i, j]
                p = Point(r, z)
                mask[i, j] *= limiter.contains(p)
        return mask

    @cached_property
    def _calc_meshes(
        self,
    ) -> tuple[tuple[NDArray[F64], NDArray[F64]], tuple[float, float, float, float]]:
        """Initialize both meshes and final extent after adjustment to achieve target resolution"""
        # Actualize the grid/mesh and update extent
        rmin, rmax, zmin, zmax = self.min_extent  # [m]
        dr, dz = self.dxgrid  # [m]
        if rmax - rmin > 0.0 and zmax - zmin > 0.0:
            rgrid = np.arange(rmin, rmax + dr, dr)  # [m] Grid that spans full extent
            zgrid = np.arange(zmin, zmax + dz, dz)  # [m]
            rmesh, zmesh = np.meshgrid(rgrid, zgrid, indexing="ij")  # [m]

            # Update the extent, which may have been adjusted
            extent = (
                float(np.min(rgrid)),
                float(np.max(rgrid)),
                float(np.min(zgrid)),
                float(np.max(zgrid)),
            )
        else:
            # If our grid spec is zero-size, make a unit mesh
            # to allow the calcs to proceed, without providing real tables
            rmesh = np.nan * np.zeros((1, 1))
            zmesh = np.nan * np.zeros((1, 1))
            extent = self.min_extent
        return (rmesh, zmesh), extent

    @cached_property
    def meshes(self) -> tuple[NDArray[F64], NDArray[F64]]:
        """[m] Shape: ((nr, nz), (nr, nz)) 2D meshgrids over domain"""
        meshes, _ = self._calc_meshes
        return meshes

    @cached_property
    def grids(self) -> tuple[NDArray[F64], NDArray[F64]]:
        """[m] with shape ((nr), (nz)) 1D grids"""
        meshes, _ = self._calc_meshes
        rgrid = meshes[0][:, 0]  # [m] all rs for first z
        zgrid = meshes[1][0, :]  # [m] all zs for first r
        return (rgrid, zgrid)

    @cached_property
    def extent(self) -> tuple[float, float, float, float]:
        """[m] The (rmin, rmax, zmin, zmax) extent of the grid cell centers"""
        _, extent = self._calc_meshes
        return extent

    @cached_property
    def extent_for_plotting(self) -> tuple[float, float, float, float]:
        """[m] The (rmin, rmax, zmin, zmax) extent of the grid cell edges"""
        rmin, rmax, zmin, zmax = self.extent
        dr, dz = self.dxgrid
        return (
            rmin - dr / 2,
            rmax + dr / 2,
            zmin - dz / 2,
            zmax + dz / 2,
        )

    @cached_property
    def nr(self) -> int:
        """Number of r-discretizations in mesh"""
        return self.grids[0].size

    @cached_property
    def nz(self) -> int:
        """Number of z-discretizations in mesh"""
        return self.grids[1].size

    @cached_property
    def n_coils(self) -> int:
        """Number of coils"""
        return len(self.coils)

    @cached_property
    def n_structures(self) -> int:
        """Number of structure filaments"""
        return len(self.structures)

    @cached_property
    def n_structure_modes(self) -> int:
        """Number of structure modes retained"""
        return len(self.structure_mode_eigenvalues)

    @cached_property
    def n_circuits(self) -> int:
        """Number of coil series circuits"""
        return len(self.circuits)

    @cached_property
    def coil_mutual_inductances(self) -> NDArray[F64]:
        """[H] with shape (ncoil X ncoil), Coil-coil mutual inductance"""
        return _calc_coil_mutual_inductances(self.coils, self.show_prog)

    @cached_property
    def structure_mutual_inductances(self) -> NDArray[F64]:
        """[H] with shape (nstruct X nstruct), Structure-structure mutual inductance"""
        return _calc_structure_mutual_inductances(self.structures, self.show_prog)

    @cached_property
    def structure_mode_mutual_inductances(self) -> NDArray[F64]:
        """[H] with shape (nmodes X nmodes), Structure mode-mode mutual inductance"""
        mss = self.structure_mutual_inductances
        tuv = self.structure_model_reduction
        return tuv.T @ (mss @ tuv)

    @cached_property
    def coil_structure_mode_mutual_inductances(self) -> NDArray[F64]:
        """[H] with shape (ncoil X nmodes), Coil-to-structure-mode mutual inductance"""
        mcs = self.coil_structure_mutual_inductances
        tuv = self.structure_model_reduction
        return mcs @ tuv

    @cached_property
    def coil_structure_mutual_inductances(self) -> NDArray[F64]:
        """[H] (ncoil X nstruct) Coil-structure mutual inductance"""
        return _calc_coil_structure_mutual_inductances(
            self.coils, self.structures, self.show_prog
        )

    @cached_property
    def circuit_mutual_inductances(self) -> NDArray[F64]:
        """[H] with shape (ncirc X ncirc), Circuit-circuit mutual inductance"""
        return _calc_circuit_mutual_inductances(
            self.circuits, self.coil_mutual_inductances, self.show_prog
        )

    @cached_property
    def circuit_structure_mutual_inductances(self) -> NDArray[F64]:
        """[H] with shape (ncirc X nstruct), Circuit-structure mutual inductance"""
        return _calc_circuit_structure_mutual_inductances(
            self.circuits, self.coil_structure_mutual_inductances, self.show_prog
        )

    @cached_property
    def circuit_structure_mode_mutual_inductances(self) -> NDArray[F64]:
        """[H] with shape (ncirc X nmodes), Circuit-to-structure-mode mutual inductance"""
        mcircs = self.circuit_structure_mutual_inductances
        tuv = self.structure_model_reduction
        return mcircs @ tuv

    @cached_property
    def coil_resistances(self) -> NDArray[F64]:
        """[ohm] with shape (ncoil X ncoil), Resistance of coils as diagonal matrix"""
        return np.diag(np.array([c.resistance for c in self.coils]))

    @cached_property
    def structure_resistances(self) -> NDArray[F64]:
        """[ohm] (nstruct X nstruct) Resistance of structure fils, diagonal"""
        return np.diag(np.array([s.resistance for s in self.structures]))

    @cached_property
    def structure_mode_resistances(self) -> NDArray[F64]:
        """[H] with shape (ncoil X nstruct), Coil-structure mutual inductance"""
        r = self.structure_resistances
        tuv = self.structure_model_reduction
        return tuv.T @ (r @ tuv)

    @cached_property
    def circuit_resistances(self) -> NDArray[F64]:
        """[ohm] with shape (ncirc X ncirc), Resistance of circuits as diagonal matrix"""
        r = np.zeros(self.n_circuits)
        # Sum the series resistance of the coils in the circuit
        for i in range(self.n_circuits):
            r[i] = np.sum(
                [self.coil_resistances[c[0], c[0]] for c in self.circuits[i].coils]
            )
        return np.diag(r)

    @cached_property
    def structure_mode_eigenvalues(self) -> NDArray[F64]:
        """
        [s] with shape (neig), eigenvalues of structure model reduction,
        kept for comparison purposes only, as these are not
        needed in order to use the structure model reduction.
        """
        d, _, _ = self._calc_model_reduction
        return d

    @cached_property
    def structure_model_reduction(self) -> NDArray[F64]:
        """
        [dimensionless] with shape (nstruct X neig), structure model-reduction matrix.\n
        Each column is a normalized vector of unit length.

        Apply this transform like `a_transformed = a @ tuv` for matrices
        with `a.shape == (..., nstruct)`.\n
        In the case of diagonalized structure filament resistances,\n
        the transform is applied like `r_transformed = tuv.T @ (r @ tuv)`.
        """
        _, tuv, _ = self._calc_model_reduction
        return tuv

    @cached_property
    def coil_flux_tables(self) -> NDArray[F64]:
        """[Wb/A] with shape (ncoils X nr X nz), Coil flux tables"""
        return _calc_coil_flux_tables(self.coils, self.meshes, self.show_prog)

    @cached_property
    def structure_flux_tables(self) -> NDArray[F64]:
        """[Wb/A] with shape (nstruct X nr X nz), Structure flux tables"""
        return _calc_structure_flux_tables(self.structures, self.meshes, self.show_prog)

    @cached_property
    def structure_mode_flux_tables(self) -> NDArray[F64]:
        """[Wb/A] with shape (nmodes X nr X nz), Structure mode flux tables"""
        return _calc_structure_mode_flux_tables(
            self.structure_flux_tables,
            self.structure_model_reduction,
            self.show_prog,
        )

    @cached_property
    def plasma_flux_tables(self) -> NDArray[F64]:
        """[Wb/A] with shape (nr*nz X nr X nz), Plasma flux tables"""
        return _calc_mesh_flux_tables(self.meshes, self.grids, self.show_prog)

    @cached_property
    def circuit_flux_tables(self) -> NDArray[F64]:
        """[Wb/A] with shape (ncirc X nr X nz), Circuit flux tables"""
        return _calc_circuit_flux_tables(
            self.circuits, self.coil_flux_tables, self.show_prog
        )

    @cached_property
    def coil_flux_density_tables(self) -> tuple[NDArray[F64], NDArray[F64]]:
        """[T/A] with shape (ncoils X nr X nz), Coil flux density (B-field) tables, r- and z- components"""
        return _calc_coil_flux_density_tables(
            self.coils, self.meshes, self.coil_flux_tables, self.show_prog
        )

    @cached_property
    def structure_flux_density_tables(self) -> tuple[NDArray[F64], NDArray[F64]]:
        """[T/A] with shape (nstruct X nr X nz), Structure flux density (B-field) tables, r- and z- components"""
        return _calc_structure_flux_density_tables(
            self.structures, self.meshes, self.structure_flux_tables, self.show_prog
        )

    @cached_property
    def structure_mode_flux_density_tables(self) -> tuple[NDArray[F64], NDArray[F64]]:
        """[T/A] with shape (nstruct X nr X nz), Structure mode flux density (B-field) tables, r- and z- components"""
        return _calc_structure_mode_flux_density_tables(
            *self.structure_flux_density_tables,
            self.structure_model_reduction,
            self.show_prog,
        )

    @cached_property
    def circuit_flux_density_tables(self) -> tuple[NDArray[F64], NDArray[F64]]:
        """[T/A] with shape (ncirc X nr X nz), Circuit flux density (B-field) tables, r- and z- components"""
        return _calc_circuit_flux_density_tables(
            self.circuits,
            *self.coil_flux_density_tables,
            self.show_prog,
        )

    @cached_property
    def coil_coil_forces(self) -> tuple[NDArray[F64], NDArray[F64]]:
        """
        [N/A^2] with shape (ncoils X ncoils), Coil-coil force tables, r- and z- components.

        Note that analytic force estimates include a variety of sources of error, including
        geometric differences between analysis and real hardware, discretization error,
        numerical summation error, and so on. Due to the importance of loads analysis, it
        is always recommended to double-check force results with at least one other tool,
        preferably of meaningfully distinct design - for example, cross-check an analytic method
        with a finite-element method.

        Because the accuracy of force estimates depends heavily on problem setup and geometry,
        no particular claims are made here about the accuracy of the force calculations,
        and they should never be used for human safety applications.
        """
        return _calc_coil_coil_forces(
            self.coils, self.grids, self.coil_flux_density_tables, self.show_prog
        )

    @cached_property
    def circuit_coil_forces(self) -> tuple[NDArray[F64], NDArray[F64]]:
        """
        [N/A^2] with shape (ncirc X ncoils), Circuit-coil force tables, r- and z- components.

        Note that analytic force estimates include a variety of sources of error, including
        geometric differences between analysis and real hardware, discretization error,
        numerical summation error, and so on. Due to the importance of loads analysis, it
        is always recommended to double-check force results with at least one other tool,
        preferably of meaningfully distinct design - for example, cross-check an analytic method
        with a finite-element method.

        Because the accuracy of force estimates depends heavily on problem setup and geometry,
        no particular claims are made here about the accuracy of the force calculations,
        and they should never be used for human safety applications.
        """
        return _calc_circuit_coil_forces(
            self.coils, self.circuits, self.coil_coil_forces, self.show_prog
        )

    @cached_property
    def structure_coil_forces(self) -> tuple[NDArray[F64], NDArray[F64]]:
        """
        [N/A^2] with shape (nstruct X ncoils), Structure filament-coil force tables, r- and z- components.

        Note that analytic force estimates include a variety of sources of error, including
        geometric differences between analysis and real hardware, discretization error,
        numerical summation error, and so on. Due to the importance of loads analysis, it
        is always recommended to double-check force results with at least one other tool,
        preferably of meaningfully distinct design - for example, cross-check an analytic method
        with a finite-element method.

        Because the accuracy of force estimates depends heavily on problem setup and geometry,
        no particular claims are made here about the accuracy of the force calculations,
        and they should never be used for human safety applications.
        """
        return _calc_structure_coil_forces(self.coils, self.structures, self.show_prog)

    @cached_property
    def structure_mode_coil_forces(self) -> tuple[NDArray[F64], NDArray[F64]]:
        """
        [N/A^2] with shape (nmodes X ncoils), Structure mode-coil force tables, r- and z- components.

        Note that analytic force estimates include a variety of sources of error, including
        geometric differences between analysis and real hardware, discretization error,
        numerical summation error, and so on. Due to the importance of loads analysis, it
        is always recommended to double-check force results with at least one other tool,
        preferably of meaningfully distinct design - for example, cross-check an analytic method
        with a finite-element method.

        Because the accuracy of force estimates depends heavily on problem setup and geometry,
        no particular claims are made here about the accuracy of the force calculations,
        and they should never be used for human safety applications.
        """
        tuv = self.structure_model_reduction
        frsc, fzsc = self.structure_coil_forces
        fr = tuv.T @ frsc
        fz = tuv.T @ fzsc
        return (fr, fz)

    @cached_property
    def plasma_coil_forces(self) -> tuple[NDArray[F64], NDArray[F64]]:
        """
        [N/A^2] with shape (nr*nz X ncoils), Mesh-coil force tables, r- and z- components.

        Note that analytic force estimates include a variety of sources of error, including
        geometric differences between analysis and real hardware, discretization error,
        numerical summation error, and so on. Due to the importance of loads analysis, it
        is always recommended to double-check force results with at least one other tool,
        preferably of meaningfully distinct design - for example, cross-check an analytic method
        with a finite-element method.

        Because the accuracy of force estimates depends heavily on problem setup and geometry,
        no particular claims are made here about the accuracy of the force calculations,
        and they should never be used for human safety applications.
        """
        if self._plasma_coil_force_method == "tables":
            plasma_flux_tables_or_limiter_mask = self.plasma_flux_tables
            full_flux_tables = True
        elif self._plasma_coil_force_method == "mask":
            plasma_flux_tables_or_limiter_mask = self.limiter_mask
            full_flux_tables = False
        else:
            raise ValueError("Unexpected calc method")

        return _calc_plasma_coil_forces(
            self.coils,
            self.grids,
            self.meshes,
            plasma_flux_tables_or_limiter_mask,
            full_flux_tables,
            show_prog=self.show_prog,
        )

    @cached_property
    def coil_filament_rzn(self) -> list[list[tuple[float, float, float]]]:
        """
        ([m], [m], [dimensionless]) r,z,n of each coil's filaments, where `n` is number of turns.
        Provided to prevent repeatedly assembling this representation when evaluating plasma inductances.
        """
        rzn = []
        for coil in self.coils:
            rzn.append([(x.r, x.z, x.n) for x in coil.filaments])
        return rzn  # [m], [m], [dimensionless]

    @cached_property
    def flux_solver(self) -> Callable[[NDArray[F64]], NDArray[F64]]:
        """
        Linear solver for extracting a flux field from a toroidal current density distribution
        using a 4th-order finite difference approximation of the Grad-Shafranov PDE.
        For `jtor` toroidal current density shaped like (nr, nz), call like `psi = flux_solver(rhs)`
        to get `psi` in [Wb] or [V-s], where `rhs = -2.0 * np.pi * mu_0 * rmesh * jtor` with the boundary
        values set to the circular-filament solved flux.
        """
        return flux_solver(self.grids)

    def get_coil_names(self) -> list[str]:
        """Get coil names in the same order as their indices"""
        # Leaving this function in for backwards compatibility
        return self.coil_names

    @cached_property
    def coil_names(self) -> list[str]:
        """Coil names in the same order as their indices"""
        return [c.name for c in self.coils]

    def get_coil_index_dict(self) -> dict[str, int]:
        """Get a map from coil names to indices in tables and self.coils"""
        # Leaving this function in for backwards compatibility
        return self.coil_index_dict

    @cached_property
    def coil_index_dict(self) -> dict[str, int]:
        """A map from coil names to indices in tables and self.coils"""
        return {self.coils[i].name: i for i in range(len(self.coils))}

    def calc_plasma_flux(
        self,
        current_density: NDArray[F64],
        calc_method: Literal["table", "solve"] = "table",
    ) -> NDArray:
        """
        Calculate the flux field associated with a given plasma current density distribution,
        either by summing over baked flux tables or by solving the Grad-Shafranov PDE.

        This calculation is most commonly used for the plasma, but is in fact more general,
        and applies to anything with an equivalent toroidal current density and axisymmetry.

        Args:
            current_density: [A/m^2], shape (nr, nz), toroidal current density on finite-difference mesh
            calc_method: Whether to sum over baked tables (memory-intensive, but reliable)
                         or perform Grad-Shafranov linear solve (lean, but fragile). Defaults to "table".

        Returns:
            poloidal flux field, [Wb] with shape (nr, nz)
        """
        assert (
            current_density.shape == self.meshes[0].shape
        ), "Supplied current density shape does not match tables"
        if calc_method == "table":
            dr, dz = self.dxgrid  # [m] grid discretization
            area = dr * dz  # [m^2] cross-sectional area of grid cell
            psi = np.zeros_like(current_density)  # [Wb]
            for i, jtor_cell in enumerate(current_density.flatten()):  # [A/m^2]
                psi_table_part = self.plasma_flux_tables[i, :, :]  # [Wb/A]
                current = jtor_cell * area  # [A]
                psi += current * psi_table_part  # [Wb]
        elif calc_method == "solve":
            psi = solve_flux_axisymmetric(
                self.grids, self.meshes, current_density, self.flux_solver
            )

        return psi  # [Wb]

    def calc_plasma_flux_density(
        self,
        plasma_flux: NDArray[F64],
    ) -> tuple[NDArray[F64], NDArray[F64]]:
        """
        Calculate a 4th-order estimate of the plasma's B-field via finite
        difference on the flux field.

        Args:
            plasma_flux: [Wb] with shape (nr, nz), Solved (or summed) plasma poloidal flux

        Returns:
            (br, bz) [T/A] with shape (nr, nz), plasma flux density (B-field)
        """
        return calc_flux_density_from_flux(plasma_flux, *self.meshes)

    def calc_plasma_self_inductance(
        self,
        plasma_current: float,
        plasma_poloidal_flux: NDArray[F64],
        br_plasma: NDArray[F64],
        bz_plasma: NDArray[F64],
        plasma_surface: tuple[NDArray[F64], NDArray[F64]],
        plasma_mask: NDArray[F64],
    ) -> tuple[float, float, float]:
        """
        Calculate the plasma's instantaneous poloidal self-inductance from two components
        addressing the energy stored inside and outside the plasma volume, returning both
        components and the total because the components are sometimes used for other calcs.

        See `self.calc_plasma_flux()`, `self.calc_plasma_flux_density()`, and
        `device_inductance.contour.trace_contour()` for methods to build the inputs
        to this function.

        This method is of marginal use on its own; because any change in current in
        the inductive system will result in motion of the plasma current distribution,
        which changes the plasma self-inductance, the instantaneous self-inductance only
        provides one of two terms in determining the loop voltage response of the plasma.

        With a system of just the plasma and examining only inductive voltage for clarity:

        ```
        V_loop = I*(dL/dt) + L*(dI/dt) = I*(dL/dI)*(dI/dt) + L*(dI/dt)
                             ^                ^
                             |                |
           This formula's L  -                - This quantity is also needed
                                                for a linear treatment
        ```

        This extends to coupled systems with multiple inductors in a similar way.

        Details of the method used can be found in the
        [docs for cfsem.](https://cfsem-py.readthedocs.io/en/latest/python/inductance/#cfsem.self_inductance_distributed_axisymmetric_conductor)

        Args:
            plasma_current: [A] total toroidal current in the plasma
            plasma_poloidal_flux: [Wb] solved plasma flux field
            br_plasma: [T] solved plasma magnetic flux density, R-component
            bz_plasma: [T] solved plasma magnetic flux density, Z-component
            plasma_surface: [m] r,z coordinates of plasma last closed flux surface (aka LCFS, aka bounding contour)
            plasma_mask: [dimensionless] binary mask of plasma interior points. 1 inside, 0 outside.

        Returns:
            (Lt, Li, Le) [H] total, internal, and external instantaneous self-inductances
        """
        total_inductance, internal_inductance, external_inductance = (
            self_inductance_distributed_axisymmetric_conductor(
                current=plasma_current,
                grid=self.grids,
                mesh=self.meshes,
                b_part=(br_plasma, bz_plasma),
                psi_part=plasma_poloidal_flux,
                mask=plasma_mask,
                edge_path=plasma_surface,
            )
        )

        return total_inductance, internal_inductance, external_inductance  # [H]

    @cached_property
    def _calc_model_reduction(self) -> tuple[NDArray[F64], NDArray[F64], int]:
        """Internal function for generating model-reduction outputs"""
        if self.model_reduction_method == "eigenmode":
            d, tuv, neig = model_reduction.eigenmode_reduction(
                self.structure_mutual_inductances,
                self.structure_resistances,
                self.max_nmodes,
            )
        elif self.model_reduction_method == "stabilized eigenmode":
            d, tuv, neig = model_reduction.stabilized_eigenmode_reduction(
                self.structure_mutual_inductances,
                self.structure_resistances,
                self.max_nmodes,
            )
        else:
            raise ValueError(
                f"Unrecognized model reduction method `{self.model_reduction_method}`"
            )

        return d, tuv, neig


@dataclass(frozen=True)
class TypicalOutputs:
    """
    A fully-actualized (precomputed) set of matrices and tables
    covering typical usage patterns.
    """

    device: DeviceInductance
    """
    Thin wrapper to provide methods and properties
    on a device description ODS.
    """

    extent: tuple[float, float, float, float]
    """
    [m] Extent of the computational domain's cell centers.
    This will be updated during initialization, during which it
    may be adjusted to satisfy the required spatial resolution.
    """
    extent_for_plotting: tuple[float, float, float, float]
    """
    [m] The full extent of the computational domain implied by
    the cell centers, expanded a half-cell to the boundary cells' edges
    like [extent[0] - dr/2, extent[1] + dr/2, ...]
    """

    dxgrid: tuple[float, float]
    """[m] spatial resolution of computational grid"""

    meshes: tuple[NDArray[F64], NDArray[F64]]
    """[m] with shape ((nr, nz), (nr, nz)), 2D meshgrids over domain"""
    grids: tuple[NDArray[F64], NDArray[F64]]
    """[m] with shape ((nr), (nz)), 1D grids"""

    nmodes: int
    """Number of structure modes retained"""

    mcc: NDArray[F64]
    """[H] with shape (ncoil X ncoil), Coil-coil mutual inductance"""
    mss: NDArray[F64]
    """[H] with shape (nstruct X nstruct), Structure-structure mutual inductance"""
    mcs: NDArray[F64]
    """[H] with shape (ncoil X nstruct), Coil-structure mutual inductance"""

    r_c: NDArray[F64]
    """[ohm] with shape (ncoil X ncoil), Resistance of coils as diagonal matrix"""
    r_s: NDArray[F64]
    """[ohm] with shape (nstruct X nstruct), Resistance of structure fils, diagonal"""
    r_modes: NDArray[F64]
    """[ohm] with shape (neig X neig), Resistance of structure modes"""

    tuv: NDArray[F64]
    """
    [dimensionless] with shape (nstruct X neig), structure model-reduction matrix.\n
    Each column is a normalized vector of unit length.

    Apply this transform like `a_transformed = a @ tuv` for matrices
    with `a.shape == (..., nstruct)`.
    In the case of diagonalized structure filament resistances,
    the transform is applied like `r_transformed = tuv.T @ (r @ tuv)`.
    """

    psi_c: NDArray[F64]
    """[Wb/A] with shape (ncoils X nr X nz), Coil flux tables"""
    psi_s: NDArray[F64]
    """[Wb/A] with shape (nstruct X nr X nz), Structure flux tables"""
    psi_modes: NDArray[F64]
    """[Wb/A] with shape (nmodes X nr X nz), Structure mode flux tables"""



import pytest

import device_inductance


@pytest.fixture(scope="session")
def typical_outputs() -> device_inductance.TypicalOutputs:
    # Set up a regular computational grid
    # It would be nice to use a coarser grid,
    # but in order for the tables to be testable,
    # we need a fairly fine one
    dxgrid = (0.05, 0.04)  # Different resolution to make sure they are never swapped
    extent = (2.0 * dxgrid[0], 4.5, -3.0, 3.0)

    # Load the default device
    ods = device_inductance.load_default_ods()

    # Pre-compute the usual set of matrices and tables
    typical_outputs = device_inductance.typical(
        ods, extent, dxgrid, max_nmodes=int(1e6), show_prog=False
    )

    return typical_outputs


@pytest.fixture(scope="session")
def typical_outputs_many_slices() -> device_inductance.TypicalOutputs:
    # Set up a regular computational grid
    # It would be nice to use a coarser grid,
    # but in order for the tables to be testable,
    # we need a fairly fine one
    dxgrid = (0.05, 0.04)  # Different resolution to make sure they are never swapped
    extent = (2.0 * dxgrid[0], 4.5, -3.0, 3.0)

    # Load the default device
    ods = device_inductance.load_default_ods()

    # Pre-compute the usual set of matrices and tables
    typical_outputs = device_inductance.typical(
        ods, extent, dxgrid, max_nmodes=int(1e6), show_prog=False, n_radial_slices=100
    )

    return typical_outputs


@pytest.fixture(scope="session")
def typical_outputs_stabilized_eigenmode() -> device_inductance.TypicalOutputs:
    """With alternate options for model reduction method and plasma-coil force method"""
    # Because the device is immutable after init, we have to make a whole new one
    # to get the other model reduction method

    # Set up a regular computational grid
    # It would be nice to use a coarser grid,
    # but in order for the tables to be testable,
    # we need a fairly fine one
    dxgrid = (0.05, 0.04)  # Different resolution to make sure they are never swapped
    extent = (2.0 * dxgrid[0], 4.5, -3.0, 3.0)

    # Load the default device
    ods = device_inductance.load_default_ods()

    # Pre-compute the usual set of matrices and tables
    typical_outputs = device_inductance.typical(
        ods,
        extent,
        dxgrid,
        max_nmodes=int(1e6),
        show_prog=False,
        model_reduction_method="stabilized eigenmode",
        plasma_coil_force_method="tables"
    )

    return typical_outputs



from dataclasses import dataclass
from warnings import warn
from functools import cached_property

import numpy as np
from numpy.typing import NDArray

from interpn import MulticubicRectilinear

from omas import ODS


@dataclass(frozen=True)
class PoloidalFieldProbe:
    """
    A poloidal B-field sensor at a point in space; aka Mirnov probe or pickup coil.
    Idealized here as attached to an ideal integrator s.t. it senses B integrated
    in time instead of rate of change.

    In reality, the hardware senses dB/dt on some vector in the poloidal plane,
    but this is not so useful in a digital context without high-bandwidth integration.
    """

    name: str
    """Sensor name"""

    r: float
    """[m] r-location"""
    phi: float
    """[rad] angular location (about z)"""
    z: float
    """[m] z-location"""

    poloidal_angle: float
    """[rad] poloidal orientation (about phi), clockwise from +R-axis"""

    @property
    def unit_vector(self) -> NDArray:
        """
        Get the unit vector in the rz plane that the measurement is projected on.
        Uses cocos-11 coordinate system, with poloidal angle clockwise about into-the-board
        toroidal direction, starting from the outboard (r=1, z=0) direction.
        """
        z = -np.sin(self.poloidal_angle)
        r = np.cos(self.poloidal_angle)

        return np.atleast_1d(np.array((r, z)))

    def response(
        self, grids: tuple[NDArray, NDArray], br: NDArray, bz: NDArray
    ) -> float:
        """Ideal integrated response to a given local B-field.  This is the sum of local B-field
        components, not the normed field!

        Args:
            grids: [m] 1D arrays describing rectilinear mesh
            br: [T]
            bz: [T]

        Raises:
            ValueError: If the sensor is not inside the mesh

        Returns:
            [T] ideal integrated B-field, sum of components projected on the axis of the probe
        """

        # Build interpolators - this is fast but not totally free
        br_interpolator = MulticubicRectilinear.new(
            [x for x in grids], br.flatten(), linearize_extrapolation=True
        )
        bz_interpolator = MulticubicRectilinear.new(
            [x for x in grids], bz.flatten(), linearize_extrapolation=True
        )

        # Unpack/repack geometry
        unit_vector = self.unit_vector
        point_to_interp = [np.atleast_1d(self.r), np.atleast_1d(self.z)]

        # Check if we are extrapolating the fields
        out_of_bounds_flags = br_interpolator.check_bounds(point_to_interp, atol=1e-3)
        if np.any(out_of_bounds_flags):
            raise ValueError(
                f"Poloidal field probe {self.name} outside of mesh bounds."
                + f" Mesh bounds flags: {out_of_bounds_flags}"
            )

        # Interpolate fields
        br_interped = br_interpolator.eval(point_to_interp)[0]  # [T]
        bz_interped = bz_interpolator.eval(point_to_interp)[0]

        # Aggregate response
        # This is the sum of B components aligned with direction of probe;
        # this is _not_ the normed magnitude, it's a sum, because the real probe integrates
        # flux components not normed field
        br_projected = br_interped * unit_vector[0]  # [T]
        bz_projected = bz_interped * unit_vector[1]
        b_projected = br_projected + bz_projected  # [T]

        # The response is based on an ideal integrator, so we get the
        # exact B-field instead of a voltage
        response = b_projected  # [T]

        return response  # [T]


@dataclass(frozen=True)
class FullFluxLoop:
    """
    Flux sensor wrapped around the entire toroid, encompassing all the
    poloidal projected area from this point inward.
    Idealized here as attached to an ideal integrator s.t. it senses
    flux integrated in time instead of rate of change.
    ::

                , - ~ ~ ~ - ,
            , '               ' ,
          ,                       ,         phi
         ,         A_pol           ,         ^
        ,                           ,        |
        ,              -----------> * p1     x
        ,                    R      ,        Z
         ,                         ,
          ,                       ,
            ,                  , '
              ' - , _ _ _ ,  '
                     ||
                    _||_
                    -  +
    """

    name: str
    """Sensor name"""

    r: float
    """[m] r-location"""
    z: float
    """[m] z-location"""

    def response(self, grids: tuple[NDArray, NDArray], psi: NDArray) -> float:
        """Ideal integrated response to a given flux field

        Args:
            grids: [m] 1D arrays describing rectilinear mesh
            psi: [Wb], 2D array of rate of change of poloidal flux

        Raises:
            ValueError: If the sensor is not inside the mesh

        Returns:
            [Wb] ideal integrated response
        """
        # Build interpolators - this is fast but not totally free
        psi_interpolator = MulticubicRectilinear.new(
            [x for x in grids], psi.flatten(), linearize_extrapolation=True
        )  # [V], from [Wb/s] or [V-s/s]

        # Unpack/repack geometry
        point_to_interp = [np.atleast_1d(self.r), np.atleast_1d(self.z)]

        # Check if we are extrapolating the fields
        out_of_bounds_flags = psi_interpolator.check_bounds(point_to_interp, atol=1e-3)
        if np.any(out_of_bounds_flags):
            raise ValueError(
                f"Full flux loop {self.name} outside of mesh bounds."
                + f" Mesh bounds flags: {out_of_bounds_flags}"
            )

        # Interpolate field
        psi_interped = psi_interpolator.eval(point_to_interp)[0]  # [Wb]

        # The response assumes an ideal integrator, so we get a flux value instead of a voltage
        response = psi_interped  # [Wb]

        return response  # [Wb]


@dataclass(frozen=True)
class PartialFluxLoop:
    """
    Flux sensor defined in terms of the opposite corners of a rectangle
    wrapped part-way around the toroid, enclosing some poloidal projected area
    but no toroidal projected area.
    Idealized here as attached to an ideal integrator s.t. it senses
    flux integrated in time instead of rate of change.

    ::

                                   Z
           * --------- * p2        ^
           |           |           |
           |           |           x ---> phi
        p1 * --------- *           R

    The normal vector is taken as clockwise from the vector formed by the first and second points.

        o (r1, z1)
        |
        |------> normal
        |
        |
        o (r0, z0)
    """

    name: str
    """Sensor name"""

    r0: float
    """[m] r-location of first corner"""
    phi0: float
    """[rad] angular location (about z) of first corner"""
    z0: float
    """[m] z-location of first corner"""

    r1: float
    """[m] r-location of second corner"""
    phi1: float
    """[rad] angular location (about z) of first corner"""
    z1: float
    """[m] z-location of second corner"""

    n_discretization: int = 1000
    """
    Number of points to use for integrating the flux field.
    Meant to be a constant that is tuned manually to target an acceptable error level.
    """

    @property
    def normal_vector(self) -> NDArray:
        """
        [dimensionless] (r, z) components of the normal vector
        of the surface enclosed by the loop
        """
        dr = self.r1 - self.r0  # [m]
        dz = self.z1 - self.z0  # [m]
        norm = (dr**2 + dz**2) ** 0.5  # [m]

        normal = np.array((dz, -dr)) / norm  # [dimensionless]
        return normal  # [dimensionless] clockwise normal vector

    @property
    def loop_frac(self) -> float:
        """Portion of the full toroid that is swept by this loop"""
        return abs(self.phi1 - self.phi0) / (2.0 * np.pi)  # [dimensionless]

    @cached_property
    def discretized_rz_section(self) -> tuple[NDArray, NDArray, NDArray]:
        """
        Generate evenly-sampled points along the r-z section of the loop and
        the projected area associated with each point.

        Because the projected area at each point is different, 1D trapezoid
        integration is inconvenient; instead, points are sampled entirely
        inside the span of the section, not including the endpoints, in order
        to support a Riemann sum over dot(B, dA) using a finer resolution.
        """
        dr = self.r1 - self.r0  # [m]
        dz = self.z1 - self.z0  # [m]
        norm = (dr**2 + dz**2) ** 0.5  # [m] length of cross-section
        length_per_segment = norm / float(self.n_discretization)  # [m]
        weights = np.linspace(0.0, 1.0, self.n_discretization + 2, endpoint=True)[
            1:-1
        ]  # [dimensionless]
        rs = self.r0 + dr * weights  # [m]
        zs = self.z0 + dz * weights  # [m]
        das = 2.0 * np.pi * rs * length_per_segment * self.loop_frac  # [m^2]

        return rs, zs, das  # [m],[m] sampled points, [m^2] projected area at each point

    def response(
        self, grids: tuple[NDArray, NDArray], br: NDArray, bz: NDArray
    ) -> float:
        """Ideal integrated response to a given local B-field

        Args:
            grids: [m] 1D arrays describing rectilinear mesh
            br: [T]
            bz: [T]

        Raises:
            ValueError: If the sensor is not inside the mesh

        Returns:
            [Wb] ideal integrated flux through the loop
        """

        # Build interpolators - this is fast but not totally free
        br_interpolator = MulticubicRectilinear.new(
            [x for x in grids], br.flatten(), linearize_extrapolation=True
        )
        bz_interpolator = MulticubicRectilinear.new(
            [x for x in grids], bz.flatten(), linearize_extrapolation=True
        )

        # Unpack/repack geometry
        rs, zs, das = self.discretized_rz_section  # [m], [m], [m^2]
        points_to_interp = [rs, zs]  # [m]
        normal = self.normal_vector  # [dimensionless]

        # Check if we are extrapolating the fields
        out_of_bounds_flags = br_interpolator.check_bounds(points_to_interp, atol=1e-3)
        if np.any(out_of_bounds_flags):
            raise ValueError(
                f"Partial flux loop {self.name} outside of mesh bounds."
                + f" Mesh bounds flags: {out_of_bounds_flags}"
            )

        # Interpolate fields
        br_interped = br_interpolator.eval(points_to_interp)  # [T]
        bz_interped = bz_interpolator.eval(points_to_interp)

        # Aggregate response
        # This is the sum of B components aligned with direction of probe;
        # this is _not_ the normed magnitude, it's a sum, because the real probe integrates
        # flux components not normed field
        br_projected = br_interped * normal[0]  # [T]
        bz_projected = bz_interped * normal[1]
        b_projected = br_projected + bz_projected  # [T]

        # The response assumes an ideal integrator, so we get the integrated flux instead of a voltage
        response = float(np.sum(b_projected * das))  # [Wb]

        return response  # [Wb]


def _extract_full_flux_loops(description: ODS) -> list[FullFluxLoop]:
    """
    Extract information about full flux loops.

    Args:
        description: Device geometric info in the format produced by device_description

    Returns:
        A list of full flux loop objects
    """
    loops = []

    ods_flux_loops = description["magnetics.flux_loop"]
    for ods_loop in ods_flux_loops.values():
        if ods_loop["type.index"] != 1:  # Type 1 is full loop
            continue
        loop = FullFluxLoop(
            name=ods_loop["name"],
            r=ods_loop["position.0.r"],
            z=ods_loop["position.0.z"],
        )
        loops.append(loop)

    return loops


def _extract_partial_flux_loops(description: ODS) -> list[PartialFluxLoop]:
    """
    Extract information about partial ("saddle") flux loops.

    Args:
        description: Device geometric info in the format produced by device_description

    Returns:
        A list of partial flux loop objects
    """
    loops = []

    ods_flux_loops = description["magnetics.flux_loop"]
    for ods_loop in ods_flux_loops.values():
        if ods_loop["type.index"] != 2:  # Type 2 is partial loop
            continue
        loop = PartialFluxLoop(
            name=ods_loop["name"],
            r0=ods_loop["position.0.r"],
            phi0=ods_loop["position.0.phi"],
            z0=ods_loop["position.0.z"],
            r1=ods_loop["position.1.r"],
            phi1=ods_loop["position.1.phi"],
            z1=ods_loop["position.1.z"],
        )
        loops.append(loop)

    return loops


def _extract_poloidal_field_probes(description: ODS) -> list[PoloidalFieldProbe]:
    """
    Extract information about poloidal field probes.

    Args:
        description: Device geometric info in the format produced by device_description

    Returns:
        A list of poloidal field probe objects
    """
    probes = []

    ods_bps = description["magnetics.b_field_pol_probe"]
    for ods_bp in ods_bps.values():
        bp_type_number = ods_bp["type.index"]
        if bp_type_number != 2:  # Type 2 is Mirnov probe
            warn(
                f"Encountered unimplemented poloidal field probe of type {bp_type_number}; skipping"
            )
            continue
        probe = PoloidalFieldProbe(
            name=ods_bp["name"],
            r=ods_bp["position.r"],
            phi=ods_bp["position.phi"],
            z=ods_bp["position.z"],
            # area_turns=ods_bp["area"],
            poloidal_angle=ods_bp["poloidal_angle"],
        )
        probes.append(probe)

    return probes



"""Top-level discretization of conducting structure."""

from __future__ import annotations

from functools import cached_property
from dataclasses import dataclass

import numpy as np
from numpy.typing import NDArray

import cfsem

from shapely import Polygon
from device_inductance import mesh

from .input import PassiveStructureInput
from .slicer import RadialSlicer
from .heuristics import poly_angle, unroundness, MAX_EDGE_LENGTH_M
from .filament import PassiveStructureFilament, _mesh_elem_to_fil


@dataclass(frozen=True)
class PassiveStructureLoop:
    """
    A logical chunk of structure material that may be the result of slicing a larger
    object into smaller pieces. Composed of one or more filaments.
    """

    # Inputs
    parent_name: str
    """Name of the input structure this was chunked from"""
    polygon: Polygon
    """Shape of the enclosing polygon before sub-discretization"""

    # Discretization results
    filaments: list[PassiveStructureFilament]  # After meshing

    @cached_property
    def rs(self) -> NDArray:
        """[m] Filament radial coordinates"""
        return np.array([f.r for f in self.filaments])

    @cached_property
    def zs(self) -> NDArray:
        """[m] Filament axial coordinates"""
        return np.array([f.z for f in self.filaments])

    @cached_property
    def ns(self) -> NDArray:
        """
        [dimensionless] (Fractional) number of turns of each filament, weighted according to their
        cross-sectional area as a fraction of this loop's total.
        """
        return np.array([f.polygon.area / self.polygon.area for f in self.filaments])

    @cached_property
    def resistance(self) -> float:
        """[ohm] Total loop resistance; effective parallel resistance over all filaments"""
        # This resistance calc treats the individual filaments as wired in parallel.
        resistance = 1.0 / sum([1.0 / f.resistance for f in self.filaments])
        return resistance  # [ohm]

    @cached_property
    def self_inductance(self) -> float:
        """[H] Self-inductance of this loop."""
        # Because the filaments within a chunk are assumed to be in parallel and isopotential on the section,
        # each one is accounted as only a fraction of a full turn - otherwise, the calculated inductance
        # would diverge as the discretization becomes finer.
        fil_frac_of_loop = self.ns
        self_inductance = 0.0  # [H]
        for i, f in enumerate(self.filaments):
            r = np.atleast_1d(f.r)
            z = np.atleast_1d(f.z)
            ref_current = fil_frac_of_loop[i]  # [A]
            mutuals = fil_frac_of_loop * cfsem.flux_circular_filament(
                ref_current, r, z, self.rs, self.zs
            )
            # Replace singularity with analytic estimate
            mutuals[i] = fil_frac_of_loop[i] ** 2 * f.self_inductance
            self_inductance += np.sum(mutuals)

        # This loop may be the result of discretizing a larger chunk of material,
        # in which case it does not represent a full loop by itself.
        # Because the self inductance is really the mutual inductance from self to self,
        # the fraction of loop needs to be accounted twice.
        # This overall fraction of loop is accounted in `self.ns` and does not need to be repeated here.
        self_inductance = float(self_inductance)

        return self_inductance  # [H]

    def mutual_inductance(self, other: PassiveStructureLoop) -> float:
        """
        [H] Mutual inductance between two loops.
        If `self` passed as `other`, the precalculated self-inductance is returned.
        """

        if id(other) == id(self):
            return self.self_inductance

        # Each loop's `ns` includes accounting of both filament number of turns and overall number of turns
        rzn1 = np.array([self.rs, self.zs, self.ns])  # [m], [m], [dimensionless]
        rzn2 = np.array([other.rs, other.zs, other.ns])

        m = cfsem.mutual_inductance_of_cylindrical_coils(rzn1, rzn2)

        return m  # [H]

    @classmethod
    def from_poly(
        cls,
        parent_name: str,
        polygon: Polygon,
        resistivity: float,
        max_edge_length_m: float = MAX_EDGE_LENGTH_M,
    ) -> PassiveStructureLoop:
        """
        Discretize a structure that is reasonably well-associated with a single centroid
        and may be the result of earlier sub-division of a larger structure, but may not be
        fully discretized yet.
        """
        # Subdivide by meshing
        sub_polygons: list[Polygon] = mesh._mesh_region(
            np.array(polygon.boundary.segmentize(max_edge_length_m).xy).T
        )

        # Make a filament from each mesh cell
        filaments = [
            _mesh_elem_to_fil(p, resistivity, parent_name) for p in sub_polygons
        ]

        # Call the collection of filaments a loop
        loop = PassiveStructureLoop(parent_name, polygon, filaments)

        return loop

    @classmethod
    def from_input(
        cls,
        inp: PassiveStructureInput,
        slicer: RadialSlicer,
        angle_thresh_deg: float = 20.0,
        unroundness_thresh: float = 2.0,
    ) -> list[PassiveStructureLoop]:
        """
        Make one or more PassiveStructureLoop from an input element.
        The input element may be subdivided into multiple loops
        if it subtends a large angle relative to the centroid
        and has a large perimeter-to-area ratio in the section plane.
        """
        # If something spans a large region around the centroid
        # AND it's not a conceptually solid block of material,
        # subdivide it so that we get adequate detail about current in different regions.
        angle_thresh_met = poly_angle(inp.polygon, slicer.centroid) > np.deg2rad(
            angle_thresh_deg
        )
        unroundness_thresh_met = unroundness(inp.polygon) > unroundness_thresh

        if angle_thresh_met and unroundness_thresh_met:
            # Slice into angular chunks
            chunks: list[Polygon] = slicer.slice(inp.polygon)
        else:
            # If the original takes up a small angular region or it's a solid block, use it as-is
            chunks: list[Polygon] = [inp.polygon]

        # Each chunk represents a fraction of one contiguous loop;
        # if we were to treat each chunk as a whole loop, the inductance of the system
        # would diverge with increasing discretization.
        # Each sub-loop's fraction of loop is weighted based on its section area.
        return [cls.from_poly(inp.parent_name, p, inp.resistivity) for p in chunks]



import numpy as np

import device_inductance

from pytest import approx

from . import typical_outputs  # Required fixture

__all__ = ["typical_outputs"]

TESTFUNC = lambda r: 1.0 / (2.0 * np.pi * r)  # noqa: E731
"""Arbitrary function for dummy field"""


def test_full_flux_loops(typical_outputs: device_inductance.TypicalOutputs):
    """Make sure the full flux loops sample the input field at a point"""
    grids = typical_outputs.grids
    rmesh, _ = typical_outputs.meshes

    testmesh = TESTFUNC(rmesh)

    assert len(typical_outputs.device.full_flux_loops) > 0
    
    for ffloop in typical_outputs.device.full_flux_loops:
        assert ffloop.response(grids, testmesh) == approx(TESTFUNC(ffloop.r), rel=1e-3)


def test_partial_flux_loops(typical_outputs: device_inductance.TypicalOutputs):
    """Make sure the partial flux loops return the integral of B*dA on their enclosed surface"""
    grids = typical_outputs.grids
    rmesh, _ = typical_outputs.meshes

    testmesh = TESTFUNC(rmesh)

    assert len(typical_outputs.device.partial_flux_loops) > 0

    unit_vector = np.array(
        [1.0 / np.sqrt(2.0), 1.0 / np.sqrt(2.0)]
    )  # direction of test field
    mag_rel = np.sqrt(2.0)  # total field magnitude relative to 1.0 / (2.0 * pi * r)
    for pfloop in typical_outputs.device.partial_flux_loops:
        # Use 1/r tendency of TESTFUNC to make flux integral easy
        dr = pfloop.r1 - pfloop.r0
        dz = pfloop.z1 - pfloop.z0
        dvec = np.array([dz, -dr])  # Perpendicular normal vector

        expected = np.dot(unit_vector, dvec) * pfloop.loop_frac * mag_rel

        response = pfloop.response(grids, testmesh, testmesh)

        assert response == approx(expected, rel=1e-3)


def test_poloidal_field_probes(typical_outputs: device_inductance.TypicalOutputs):
    """Make sure the pickup coils return the sum of projected components of the B-field on their axis"""
    grids = typical_outputs.grids
    rmesh, zmesh = typical_outputs.meshes

    assert len(typical_outputs.device.poloidal_field_probes) > 0

    # Field varying on R
    testmesh = TESTFUNC(rmesh)
    for bpcoil in typical_outputs.device.poloidal_field_probes:
        expected = TESTFUNC(bpcoil.r) * bpcoil.unit_vector[0]
        response = bpcoil.response(grids, testmesh, np.zeros_like(testmesh))
        assert response == approx(expected, rel=1e-3)

    # Field varying on Z
    testmesh = zmesh  # Avoid div/0 with testfunc
    for bpcoil in typical_outputs.device.poloidal_field_probes:
        expected = bpcoil.z * bpcoil.unit_vector[1]
        response = bpcoil.response(grids, np.zeros_like(testmesh), testmesh)
        assert response == approx(expected, rel=1e-3)

    # Field varying on R and Z
    for bpcoil in typical_outputs.device.poloidal_field_probes:
        expected = (
            bpcoil.z * bpcoil.unit_vector[1]
            + TESTFUNC(bpcoil.r) * bpcoil.unit_vector[0]
        )
        response = bpcoil.response(grids, TESTFUNC(rmesh), zmesh)
        assert response == approx(expected, rel=1e-3)



from itertools import product

import numpy as np
from numpy.typing import NDArray

from device_inductance.coils import Coil
from device_inductance.circuits import CoilSeriesCircuit
from device_inductance.structures import PassiveStructureLoop
from device_inductance.utils import _progressbar

from cfsem import mutual_inductance_of_cylindrical_coils, flux_circular_filament


def _calc_coil_mutual_inductances(coils: list[Coil], show_prog: bool = True) -> NDArray:
    """Populate coil-coil mutual inductance matrix"""

    ncoil = len(coils)
    mcc = np.zeros((ncoil, ncoil))  # [H]
    #   Coil self-inductances
    for i, c in enumerate(coils):
        mcc[i, i] = c.self_inductance
    #   Coil mutual inductances
    items = list(product(enumerate(coils), enumerate(coils)))
    if show_prog:
        items = _progressbar(items, "Coil-coil mutual inductances", show_every=ncoil)
    for (i, c1), (j, c2) in items:
        if i == j:
            # This is a self-inductance term and was already populated
            continue
        if mcc[i, j] != 0.0 and mcc[j, i] != 0.0:
            # We already populated this one from the other direction
            continue

        elems_1 = c1.filaments
        elem_rs_1 = [x.r for x in elems_1]
        elem_zs_1 = [x.z for x in elems_1]
        elem_ns_1 = [x.n for x in elems_1]
        fils1 = np.vstack((elem_rs_1, elem_zs_1, elem_ns_1))

        elems_2 = c2.filaments
        elem_rs_2 = [x.r for x in elems_2]
        elem_zs_2 = [x.z for x in elems_2]
        elem_ns_2 = [x.n for x in elems_2]
        fils2 = np.vstack((elem_rs_2, elem_zs_2, elem_ns_2))

        mcc[i, j] = mutual_inductance_of_cylindrical_coils(fils1, fils2)
        mcc[j, i] = mcc[i, j]  # Mutual inductance is reflexive

    return np.ascontiguousarray(mcc)  # [H]


def _calc_structure_mutual_inductances(
    structures: list[PassiveStructureLoop], show_prog: bool = True
) -> NDArray:
    # Populate passive system mutual inductance matrix
    n = len(structures)
    mss = np.zeros((n, n))
    for i in range(n):
        for j in range(n):
            mss[i, j] = structures[i].mutual_inductance(structures[j])

    return np.ascontiguousarray(mss)  # [H]


def _calc_coil_structure_mutual_inductances(
    coils: list[Coil],
    structures: list[PassiveStructureLoop],
    show_prog: bool = True,
) -> NDArray:
    # Populate coil-passive interaction mutual inductance matrix
    ncoil = len(coils)
    npassive = len(structures)

    mcs = np.zeros((ncoil, npassive))  # [H]

    items = [x for x in enumerate(coils)]
    if show_prog:
        items = _progressbar(items, "Coil-passive mutual inductances")
    for i, c in items:
        for j, s in enumerate(structures):
            # Use number of turns as filament current to capture (nturns * [unit current])
            mcs[i, j] = np.sum(s.ns * flux_circular_filament(c.ns, c.rs, c.zs, s.rs, s.zs))

    return np.ascontiguousarray(mcs)  # [H]


def _calc_circuit_mutual_inductances(
    circuits: list[CoilSeriesCircuit],
    coil_mutual_inductances: NDArray,
    show_prog: bool = True,
) -> NDArray:
    # Populate circuit-circuit mutual inductance matrix from coils
    ncirc = len(circuits)

    # Wiring direction affects mutual inductances but not self-inductances,
    # so it's convenient to work with a modified version of the coil-coil matrix
    mcc_signed = coil_mutual_inductances.copy()
    for circ in circuits:
        for i, sign in circ.coils:
            mcc_signed[i, :] *= sign
            mcc_signed[:, i] *= sign
            # Self inductance is still positive even if sign is negative
            # Any two coils wired in reverse will have their mutual inductance
            # sign cancel out and become positive again

    m = np.zeros((ncirc, ncirc))  # [H]
    items = [x for x in enumerate(circuits)]
    if show_prog:
        items = _progressbar(items, "Circuit-circuit mutual inductances")
    for i, circi in items:
        icoilinds = [c[0] for c in circi.coils]
        for j, circj in enumerate(circuits):
            # This procedure works for both self- and mutual- terms
            jcoilinds = [c[0] for c in circj.coils]
            m[i, j] = np.sum(
                mcc_signed[icoilinds, :][:, jcoilinds]
            )  # slice rows then cols
            m[j, i] = m[i, j]

    return m  # [H]


def _calc_circuit_structure_mutual_inductances(
    circuits: list[CoilSeriesCircuit],
    coil_structure_mutual_inductances: NDArray,
    show_prog: bool = True,
) -> NDArray:
    ncirc = len(circuits)
    nstruct = coil_structure_mutual_inductances.shape[1]

    m = np.zeros((ncirc, nstruct))  # [H]
    items = [x for x in enumerate(circuits)]
    if show_prog:
        items = _progressbar(items, "Circuit-structure mutual inductances")
    for i, circ in items:
        for j, sign in circ.coils:
            m[i, :] += sign * coil_structure_mutual_inductances[j, :]

    return m  # [H]



"""Calculation of forces between current-carrying conductors"""

import numpy as np
from numpy.typing import NDArray

from device_inductance.coils import Coil
from device_inductance.structures import PassiveStructureLoop
from device_inductance.circuits import CoilSeriesCircuit
from device_inductance.utils import _progressbar, calc_flux_density_from_flux
from device_inductance.logging import log

from cfsem import (
    flux_density_circular_filament,
    body_force_density_circular_filament_cartesian,
)

from interpn import MulticubicRectilinear


def _calc_coil_coil_forces(
    coils: list[Coil],
    grids: tuple[NDArray, NDArray],
    coil_flux_density_tables: tuple[NDArray, NDArray],
    show_prog: bool = True,
) -> tuple[NDArray, NDArray]:
    ncoils = len(coils)
    fr = np.zeros((ncoils, ncoils))  # [N/A^2]
    fz = np.zeros((ncoils, ncoils))
    gridlist = [x for x in grids]

    # Calculate force per amp from each coil `i` to each coil `j`
    # using the baked tables, which include the self-field solve patch
    # when it is available (for coils that fall on a regular grid)
    items = (
        _progressbar([x for x in range(ncoils)], "Coil-coil force rows")
        if show_prog
        else range(ncoils)
    )
    for i in items:
        bz = coil_flux_density_tables[1][i, :, :]
        bz_interp = MulticubicRectilinear.new(gridlist, bz)
        for j in range(ncoils):
            if i == j and coils[i].local_fields is None:
                # If we can't make a sane self-field estimate, skip and issue a warning
                log().warning(
                    f"Skipping self-force contribution for coil {coils[i].name} due to lack of smooth local field approximation"
                )
                continue
            elif i == j:
                # If we're doing self-field and we have a local field solve, interpolate on the
                # smooth local field to resolve the singularity issue

                # Integral of I*cross(dL,B)/I with dL in +phi direction = 2*pi*r * nturns * (Bz, 0.0, -Br)
                length_factor = 2.0 * np.pi * coils[j].rs * coils[j].ns  # [m]-turns
                obs = [
                    coils[j].rs,
                    coils[j].zs,
                ]  # [m] observation points (filament locations)
                fr[i][j] = np.sum(length_factor * bz_interp.eval(obs))
                fz[i][j] = (
                    0.0  # No self-propulsion; interpolation would produce slightly nonzero value
                )
            else:
                # If these are two separate coils, we can use a full IxB calc
                # which is slower but more accurate than interpolation

                coila = coils[i]
                coilb = coils[j]

                ra, za, na = coila.rs, coila.zs, coila.ns
                rb, zb, nb = coilb.rs, coilb.zs, coilb.ns
                zero = np.zeros_like(rb)
                # Replacing J with I*dL gives body force instead of body force density
                # and we can use the full circular length to scale the I*dL product in the toroidal direction
                fab_jxb_r, fab_jxb_y, fab_jxb_z = (
                    body_force_density_circular_filament_cartesian(
                        na,
                        ra,
                        za,
                        obs=(rb, zero, zb),
                        j=(zero, 2.0 * np.pi * rb * nb, zero),
                    )
                )  # [N/A^2]
                assert sum(fab_jxb_y) == 0.0  # Sanity check
                # Sum contributions at each filament
                fr[i][j] = sum(fab_jxb_r)
                fz[i][j] = sum(fab_jxb_z)

    return (fr, fz)


def _calc_circuit_coil_forces(
    coils: list[Coil],
    circuits: list[CoilSeriesCircuit],
    coil_coil_forces: tuple[NDArray, NDArray],
    show_prog: bool = True,
) -> tuple[NDArray, NDArray]:
    ncirc = len(circuits)
    ncoils = len(coils)
    fr = np.zeros((ncirc, ncoils))  # [N/A^2]
    fz = np.zeros((ncirc, ncoils))

    # Calculate force per amp from each circuit `i` to each coil `j` using coil-coil force tables
    items = (
        _progressbar([x for x in range(ncirc)], "Circuit-coil force rows")
        if show_prog
        else range(ncirc)
    )
    for i in items:
        for j, sign in circuits[i].coils:
            # For each coil in the circuit, add the signed force from that coil
            # on each of the others.
            # If any coils do not have self-force estimates, that will be handled earlier in the coil-coil force tables.
            fr[i, :] += sign * coil_coil_forces[0][j, :]
            fz[i, :] += sign * coil_coil_forces[1][j, :]

    return (fr, fz)


def _calc_structure_coil_forces(
    coils: list[Coil],
    structures: list[PassiveStructureLoop],
    show_prog: bool = True,
) -> tuple[NDArray, NDArray]:
    ncoils = len(coils)
    nstruct = len(structures)

    fr = np.zeros((nstruct, ncoils))  # [N/A^2]
    fz = np.zeros((nstruct, ncoils))
    items = (
        _progressbar(structures, "Structure-coil force rows")
        if show_prog
        else structures
    )
    for i, s in enumerate(items):
        ifil = s.ns  # [dimensionless] unit reference current for normalization times number of turns
        rfil = s.rs  # [m]
        zfil = s.zs  # [m]
        for j in range(ncoils):
            r = coils[j].rs  # [m]
            z = coils[j].zs  # [m]
            n = coils[j].ns  # [dimensionless]
            length_factor = 2.0 * np.pi * r * n
            zero = np.zeros_like(r)

            # Replacing J with I*dL gives body force instead of body force density
            # and we can use the full circular length to scale the I*dL product in the toroidal direction
            frij, _, fzij = body_force_density_circular_filament_cartesian(
                ifil,
                rfil,
                zfil,
                obs=(r, zero, z),
                j=(zero, length_factor, zero),
                par=False,
            )
            fr[i, j] = sum(frij)
            fz[i, j] = sum(fzij)

    return (fr, fz)


def _calc_plasma_coil_forces(
    coils: list[Coil],
    grids: tuple[NDArray, NDArray],
    meshes: tuple[NDArray, NDArray],
    plasma_flux_tables_or_limiter_mask: NDArray,
    full_flux_tables: bool = False,
    show_prog: bool = True,
) -> tuple[NDArray, NDArray]:
    ncoil = len(coils)
    nr, nz = (len(grids[0]), len(grids[1]))
    nrnz = nr * nz
    rmesh, zmesh = meshes
    fr = np.zeros((nrnz, ncoil))
    fz = np.zeros((nrnz, ncoil))
    gridlist = [x for x in grids]

    if full_flux_tables:
        # We're using the full tables
        plasma_flux_tables = plasma_flux_tables_or_limiter_mask
        items = (
            _progressbar(
                [x for x in range(nrnz)], "Mesh cell-coil force rows", show_every=nr
            )
            if show_prog
            else range(nrnz)
        )
        for i in items:
            br, bz = calc_flux_density_from_flux(
                plasma_flux_tables[i, :, :], *meshes
            )  # [T/A]
            br_interp = MulticubicRectilinear.new(gridlist, br)  # [T/A] vs. [m]
            bz_interp = MulticubicRectilinear.new(gridlist, bz)

            for j in range(ncoil):
                # Integral of I*cross(dL,B)/I with dL in +phi direction = 2*pi*r * nturns * (Bz, 0.0, -Br)
                length_factor = 2.0 * np.pi * coils[j].rs * coils[j].ns  # [m]-turns
                obs = [
                    coils[j].rs,
                    coils[j].zs,
                ]  # [m] observation points (filament locations)
                fr[i][j] = np.sum(length_factor * bz_interp.eval(obs))
                fz[i][j] = np.sum(-length_factor * br_interp.eval(obs))
    else:
        # We're using the limiter mask, so we can calculate contributions directly,
        # visiting only the points on the interior of the limiter and leaving the others as zeroes
        limiter_mask = plasma_flux_tables_or_limiter_mask.flatten()
        items = (
            _progressbar(
                [x for x in range(nrnz)], "Mesh cell-coil force rows", show_every=nr
            )
            if show_prog
            else range(nrnz)
        )
        current = np.atleast_1d(np.ones(1))
        for i in items:
            if limiter_mask[i]:
                for j in range(ncoil):
                    length_factor = 2.0 * np.pi * coils[j].rs * coils[j].ns  # [m]-turns
                    rprime, zprime = [
                        coils[j].rs,
                        coils[j].zs,
                    ]  # [m] observation points (filament locations)
                    rfil = np.atleast_1d(rmesh.flatten()[i])
                    zfil = np.atleast_1d(zmesh.flatten()[i])
                    brs, bzs = flux_density_circular_filament(
                        current, rfil, zfil, rprime, zprime, par=False
                    )  # [T/A], not enough points to benefit from parallel impl
                    fr[i][j] = np.sum(length_factor * bzs)
                    fz[i][j] = np.sum(-length_factor * brs)

    return fr, fz



import numpy as np

import matplotlib.pyplot as plt

import device_inductance
from device_inductance.utils import calc_flux_density_from_flux
from device_inductance.contour import trace_contour

from cfsem import self_inductance_circular_ring_wien

from pytest import approx

from interpn import MulticubicRegular

from . import typical_outputs  # Required fixture

__all__ = ["typical_outputs"]


def test_coil_tables(typical_outputs: device_inductance.TypicalOutputs):
    """
    Test flux tables by using them to back-calculate off-diagonal
    mutual inductance terms.

    Test B-field tables against finite difference calc.
    """
    rtol = 2e-3  # Flux error is mostly from grid resolution
    atol_b = 3e-8  # [T/A-turn] small values see some numerical error
    psi = typical_outputs.psi_c  # [Wb/A]
    br, bz = typical_outputs.device.coil_flux_density_tables  # [T/A]
    mcc = typical_outputs.mcc  # [H]

    coils = typical_outputs.device.coils
    ncoils = len(coils)

    # Interpolator inputs
    dims = np.array(psi[0, :, :].shape)
    starts = np.array([x[0] for x in typical_outputs.grids])
    steps = np.array(typical_outputs.dxgrid)

    # Test flux tables against mutual inductance matrix
    for i in range(ncoils):
        psi_interp = MulticubicRegular.new(dims, starts, steps, psi[i, :, :])
        for j in range(ncoils):
            if j == i:
                continue  # Skip self-inductance terms
            # Repack target filament locations
            robs = np.array([x.r for x in coils[j].filaments])  # [m]
            zobs = np.array([x.z for x in coils[j].filaments])  # [m]
            nobs = np.array([x.n for x in coils[j].filaments])  # target filament turns

            mutual_inductance_interped = np.sum(nobs * psi_interp.eval([robs, zobs]))
            assert mutual_inductance_interped == approx(mcc[i, j], rel=rtol)

        # Calculate B-field by alternative method
        br_from_psi, bz_from_psi = calc_flux_density_from_flux(
            psi[i, :, :], *typical_outputs.meshes
        )

        # Check B-field inside the limiter, where the results affect shaping,
        # and where the results are not affected by filaments that come too close
        # to grid points
        mask = typical_outputs.device.limiter_mask
        assert np.allclose(br[i, :, :] * mask, br_from_psi * mask, rtol, atol_b)
        assert np.allclose(bz[i, :, :] * mask, bz_from_psi * mask, rtol, atol_b)

        limiter_path = typical_outputs.device.limiter.boundary.xy

        _fig, axes = plt.subplots(2, 4, figsize=(14, 6), sharex=True, sharey=True)
        axes = np.array(axes).flatten()
        plt.sca(axes[0])
        plt.plot(*limiter_path, color="k")
        plt.imshow(
            ((br[i, :, :] - br_from_psi) / br[i, :, :]).T,
            origin="lower",
            extent=typical_outputs.extent,
        )
        plt.colorbar()
        plt.title("br rel err")

        plt.sca(axes[1])
        plt.plot(*limiter_path, color="k")
        plt.imshow(
            (br[i, :, :] - br_from_psi).T, origin="lower", extent=typical_outputs.extent
        )
        plt.colorbar()
        plt.title("br err")

        plt.sca(axes[2])
        plt.plot(*limiter_path, color="k")
        plt.imshow(br_from_psi.T, origin="lower", extent=typical_outputs.extent)
        plt.colorbar()
        plt.title("br from psi")

        plt.sca(axes[3])
        plt.plot(*limiter_path, color="k")
        plt.imshow(br[i, :, :].T, origin="lower", extent=typical_outputs.extent)
        plt.colorbar()
        plt.title("br from filaments")

        plt.sca(axes[4])
        plt.plot(*limiter_path, color="k")
        plt.imshow(
            ((bz[i, :, :] - bz_from_psi) / bz[i, :, :]).T,
            origin="lower",
            extent=typical_outputs.extent,
        )
        plt.colorbar()
        plt.title("bz rel err")

        plt.sca(axes[5])
        plt.plot(*limiter_path, color="k")
        plt.imshow(
            (bz[i, :, :] - bz_from_psi).T, origin="lower", extent=typical_outputs.extent
        )
        plt.colorbar()
        plt.title("bz err")

        plt.sca(axes[6])
        plt.plot(*limiter_path, color="k")
        plt.imshow(bz_from_psi.T, origin="lower", extent=typical_outputs.extent)
        plt.colorbar()
        plt.title("bz from psi")

        plt.sca(axes[7])
        plt.plot(*limiter_path, color="k")
        plt.imshow(bz[i, :, :].T, origin="lower", extent=typical_outputs.extent)
        plt.colorbar()
        plt.title("bz from filaments")

        plt.suptitle(coils[i].name)
        plt.tight_layout()

    plt.show()


def test_structure_tables(typical_outputs: device_inductance.TypicalOutputs):
    """
    Test flux tables by using them to back-calculate off-diagonal
    mutual inductance terms.
    """
    rtol = 1e-2  # Error is mostly from grid resolution
    psi_s = typical_outputs.psi_s  # [Wb/A]
    br, bz = typical_outputs.device.structure_flux_density_tables  # [T/A]
    mss = typical_outputs.mss  # [H]

    structures = typical_outputs.device.structures
    nstructs = len(structures)

    # interpolator inputs
    dims = [x for x in psi_s[0, :, :].shape]
    starts = np.array([x[0] for x in typical_outputs.grids])
    steps = np.array(typical_outputs.dxgrid)

    for i in range(nstructs):
        psi_interp = MulticubicRegular.new(dims, starts, steps, psi_s[i, :, :])
        # r = structures[i].rs
        # z = structures[i].zs
        for j in range(nstructs):
            if j == i:
                continue  # Skip self-inductance terms

            robs = structures[j].rs
            zobs = structures[j].zs

            mutual_inductance_interped = np.sum(structures[j].ns * psi_interp.eval([robs, zobs]))
            assert mutual_inductance_interped == approx(mss[i, j], rel=rtol)

        # For the structure's effect on B-field, we mostly care about the interior
        # of the limiter, where it affects plasma shaping, but we can test it everywhere
        meshes = typical_outputs.meshes
        br_from_psi, bz_from_psi = calc_flux_density_from_flux(psi_s[i, :, :], *meshes)

        assert np.allclose(br[i, :, :], br_from_psi, rtol=1e-3, atol=3e-8)
        assert np.allclose(bz[i, :, :], bz_from_psi, rtol=1e-3, atol=3e-8)


def test_plasma_tables_solver_inductance(
    typical_outputs: device_inductance.TypicalOutputs,
):
    """
    Compare plasma flux tables to results of a grad-shafranov solve, and compare
    plasma self-inductance to the circular ring formula.
    """

    rtol = 2e-3

    # Use the test mesh for the G-S solve for simplicity
    rmesh, zmesh = typical_outputs.meshes
    dr, dz = typical_outputs.dxgrid

    # Make a mockup of a current density profile
    ref_current = 1e3 * np.sqrt(
        np.e
    )  # [A] some number not an integer or multiple of pi
    minor_radius = 0.4  # [m] some number inside limiter
    r0 = 1.8
    z0 = 0.0
    dl = ((rmesh - r0) ** 2 + (zmesh - z0) ** 2) ** 0.5
    jtor_smooth = 1.0 / (1.0 + dl**2)
    jtor = jtor_smooth.copy()
    jtor[np.where(dl > minor_radius)] = 0.0
    jtor *= ref_current / (np.sum(jtor) * dr * dz)

    # Get plasma flux field by summing tables and by doing a grad-shafranov solve
    psi_tabulated = typical_outputs.device.calc_plasma_flux(
        current_density=jtor, calc_method="table"
    )
    psi_gs = typical_outputs.device.calc_plasma_flux(
        current_density=jtor, calc_method="solve"
    )

    # Compare tabulated flux summation to G-S solve
    assert np.allclose(psi_tabulated, psi_gs, rtol=rtol)

    # Plasma tables are already made by this method so it's a little redundant,
    # but does help to make sure the flux gradient is not too different between
    # the two methods and that the plasma flux calc passthrough runs at all
    br_plasma_1, bz_plasma_1 = typical_outputs.device.calc_plasma_flux_density(
        psi_tabulated
    )
    br_plasma_2, bz_plasma_2 = typical_outputs.device.calc_plasma_flux_density(psi_gs)
    assert np.allclose(br_plasma_1, br_plasma_2, rtol=rtol, atol=1e-8 * ref_current)
    assert np.allclose(bz_plasma_1, bz_plasma_2, rtol=rtol, atol=1e-8 * ref_current)

    # Check self-inductance against circular-section self-inductance calc
    #     Trace boundary
    rstart = np.max(rmesh[np.where(jtor > 0.0)])
    contour = trace_contour(
        typical_outputs.grids,
        jtor_smooth,  # Use jtor because we never formulated a consistent total flux field
        (rstart, z0),
        (r0, z0),
        30.0,
        typical_outputs.device.limiter_mask,
        ds=1e-3,
        tol=1e-6,
    )
    assert contour is not None
    #    Calculate self-inductance
    plasma_self_inductance, _, _ = typical_outputs.device.calc_plasma_self_inductance(
        ref_current,
        psi_gs,
        br_plasma_2,
        bz_plasma_2,
        contour,
        np.where(jtor > 0.0, 1.0, 0.0),
    )  # [H]
    approx_plasma_self_inductance = self_inductance_circular_ring_wien(
        r0, minor_radius
    )  # [H]

    #    The result won't match exactly because this is not a uniform current density distribution,
    #    but it should be fairly close
    assert plasma_self_inductance == approx(approx_plasma_self_inductance, rel=0.15)

    # Plot flux fields and error
    _fig, axes = plt.subplots(1, 4, figsize=(14, 4))
    plt.sca(axes[0])
    plt.imshow(jtor.T, extent=typical_outputs.extent_for_plotting, origin="lower")
    plt.plot(*contour, color="b")
    plt.title("jtor [A/m^2]")

    plt.sca(axes[1])
    plt.imshow(psi_gs.T, extent=typical_outputs.extent_for_plotting, origin="lower")
    plt.plot(*contour, color="b")
    plt.title("Solved flux from G-S [Wb]")

    plt.sca(axes[2])
    plt.imshow(
        psi_tabulated.T, extent=typical_outputs.extent_for_plotting, origin="lower"
    )
    plt.plot(*contour, color="b")
    plt.title("Summed flux from tables [Wb]")

    plt.sca(axes[3])
    plt.imshow(
        (psi_tabulated.T - psi_gs.T) / psi_gs.T,
        extent=typical_outputs.extent_for_plotting,
        origin="lower",
    )
    plt.colorbar()
    plt.plot(*contour, color="b")
    plt.title("Error []")

    plt.tight_layout()

    plt.show()
